r"""
# XYMap

The `XYMap` class represents one XY-grid of interconnected map-legend components. It's built from an
ASCII representation, where unique characters represents each type of component. The Map parses the
map into an internal graph that can be efficiently used for pathfinding the shortest route between
any two nodes (rooms).

Each room (MapNode) can have exits (links) in 8 cardinal directions (north, northwest etc) as well
as up and down. These are indicated in code as 'n', 'ne', 'e', 'se', 's', 'sw', 'w',
'nw', 'u' and 'd'.


```python
    # in module passed to 'Map' class

    MAP = r'''
                           1
     + 0 1 2 3 4 5 6 7 8 9 0

    10 #   # # #     #-I-#
        \  i i i     d
     9   #-#-#-#     |
         |\    |     u
     8   #-#-#-#-----#b----o
         |     |           |
     7   #-#---#-#-#-#-#   |
         |         |x|x|   |
     6   o-#-#-#   #-#-#-#b#
            \      |x|x|
     5   o---#-#<--#-#-#
        /    |
     4 #-----+-# #---#
        \    | |  \ /
     3   #b#-#-#   x   #
             | |  / \ u
     2       #-#-#---#
             ^       d
     1       #-#     #
             |
     0 #-#---o

     + 0 1 2 3 4 5 6 7 8 9 1
                           0

    '''


    LEGEND = {'#': xyzgrid.MapNode, '|': xyzgrid.NSMapLink,...}

    # read by parser if XYMAP_DATA_LIST doesn't exist
    XYMAP_DATA = {
        "map": MAP,
        "legend": LEGEND,
        "zcoord": "City of Foo",
        "prototypes": {
            (0,1): { ... },
            (1,3): { ... },
            ...
        }

    }

    # will be parsed first, allows for multiple map-data dicts from one module
    XYMAP_DATA_LIST = [
        XYMAP_DATA
    ]

```

The two `+` signs  in the upper/lower left corners are required and marks the edge of the map area.
The origo of the grid is always two steps right and two up from the bottom test marker and the grid
extends to two lines below the top-left marker. Anything outside the grid is ignored, so numbering
the coordinate axes is optional but recommended for readability.

The XY positions represent coordinates positions in the game world. When existing, they are usually
represented by Rooms in-game. The links between nodes would normally represent Exits, but the length
of links on the map have no in-game equivalence except that traversing a multi-step link will place
you in a location with an XY coordinate different from what you'd expect by a single step (most
games don't relay the XY position to the player anyway).

In the map string, every XY coordinate must have exactly one spare space/line between them - this is
used for node linkings. This finer grid which has 2x resolution of the `XYgrid` is only used by the
mapper and is referred to as the `xygrid` (small xy) internally. Note that an XY position can also
be held by a link (for example a passthrough).

The nodes and links can be customized by add your own implementation of `MapNode` or `MapLink` to
the LEGEND dict, mapping them to a particular character symbol. A `MapNode` can only be added
on an even XY coordinate while `MapLink`s can be added anywhere on the xygrid.

See `./example.py` for a full grid example.

----
"""
import pickle
from collections import defaultdict
from os import mkdir
from os.path import isdir, isfile
from os.path import join as pathjoin

try:
    from scipy import zeros
    from scipy.sparse import csr_matrix
    from scipy.sparse.csgraph import dijkstra
except ImportError as err:
    raise ImportError(
        f"{err}\nThe XYZgrid contrib requires "
        "the SciPy package. Install with `pip install scipy'."
    )
from django.conf import settings

from evennia.prototypes import prototypes as protlib
from evennia.prototypes.spawner import flatten_prototype
from evennia.utils import logger
from evennia.utils.utils import is_iter, mod_import, variable_from_module

from . import xymap_legend
from .utils import BIGVAL, MapError, MapParserError

_NO_DB_PROTOTYPES = True
if hasattr(settings, "XYZGRID_USE_DB_PROTOTYPES"):
    _NO_DB_PROTOTYPES = not settings.XYZGRID_USE_DB_PROTOTYPES

_CACHE_DIR = settings.CACHE_DIR
_LOADED_PROTOTYPES = None
_XYZROOMCLASS = None

MAP_DATA_KEYS = ["zcoord", "map", "legend", "prototypes", "options", "module_path"]

DEFAULT_LEGEND = xymap_legend.LEGEND

# --------------------------------------------
# Map parser implementation


class XYMap:
    r"""
    This represents a single map of interconnected nodes/rooms, parsed from a ASCII map
    representation.

    Each room is connected to each other as a directed graph with optional 'weights' between the the
    connections. It is created from a map string with symbols describing the topological layout. It
    also provides pathfinding using the Dijkstra algorithm.

    The map-string is read from a string or from a module.  The grid area of the string is marked by
    two `+` characters - one in the top left of the area and the other in the bottom left.
    The grid starts two spaces/lines in from the 'open box' created by these two markers and extend
    any width to the right.
    Any other markers or comments can be added outside of the grid - they will be ignored.  Every
    grid coordinate must always be separated by exactly one space/line since the space between
    are used for links.
    ::
        '''
                               1 1 1
         + 0 1 2 3 4 5 6 7 8 9 0 1 2 ...

         4       #         #   #
                 |          \ /
         3     #-#-#     #   #
               |          \ /
         2     #-#-#       #
               |x|x|       |
         1     #-#-#-#-#-#-#
              /
         0 #-#

         + 0 1 2 3 4 5 6 7 8 9 1 1 1 ...
                               0 1 2
        '''

    So origo (0,0) is in the bottom-left and north is +y movement, south is -y movement
    while east/west is +/- x movement as expected. Adding numbers to axes is optional
    but recommended for readability!

    """
    mapcorner_symbol = "+"
    max_pathfinding_length = 500
    empty_symbol = " "
    # we normally only accept one single character for the legend key
    legend_key_exceptions = "\\"

    def __init__(self, map_module_or_dict, Z="map", xyzgrid=None):
        """
        Initialize the map parser by feeding it the map.

        Args:
            map_module_or_dict (str, module or dict): Path or module pointing to a map. If a dict,
                this should be a dict with a MAP_DATA key 'map' and optionally a 'legend'
                dicts to specify the map structure.
            Z (int or str, optional): Name or Z-coord for for this map. Needed if the game uses
                more than one map. If not given, it can also be embedded in the
                `map_module_or_dict`. Used when referencing this map during map transitions,
                baking of pathfinding matrices etc.
            xyzgrid (.xyzgrid.XYZgrid): A top-level grid this map is a part of.

        Notes:
            Interally, the map deals with two sets of coordinate systems:
            - grid-coordinates x,y are the character positions in the map string.
            - world-coordinates X,Y are the in-world coordinates of nodes/rooms.
              There are fewer of these since they ignore the 'link' spaces between
              the nodes in the grid, s

                  X = x // 2
                  Y = y // 2

            - The Z-coordinate, if given, is only used when transitioning between maps
              on the supplied `grid`.

        """
        global _LOADED_PROTOTYPES
        if not _LOADED_PROTOTYPES:
            # inject default prototypes, but don't override prototype-keys loaded from
            # settings, if they exist (that means the user wants to replace the defaults)
            protlib.load_module_prototypes(
                "evennia.contrib.grid.xyzgrid.prototypes", override=False
            )
            _LOADED_PROTOTYPES = True

        self.Z = Z
        self.xyzgrid = xyzgrid

        self.mapstring = ""
        self.raw_mapstring = ""

        # store so we can reload
        self.map_module_or_dict = map_module_or_dict

        self.prototypes = None
        self.options = None

        # transitional mapping
        self.symbol_map = None

        # map setup
        self.xygrid = None
        self.XYgrid = None
        self.display_map = None
        self.max_x = 0
        self.max_y = 0
        self.max_X = 0
        self.max_Y = 0

        # Dijkstra algorithm variables
        self.node_index_map = None
        self.dist_matrix = None
        self.pathfinding_routes = None

        self.pathfinder_baked_filename = None
        if Z:
            if not isdir(_CACHE_DIR):
                mkdir(_CACHE_DIR)
            self.pathfinder_baked_filename = pathjoin(_CACHE_DIR, f"{Z}.P")

        # load data and parse it
        self.reload()

    def __str__(self):
        """
        Print the string representation of the map.
        Since the y-axes origo is at the bottom, we must flip the
        y-axis before printing (since printing is always top-to-bottom).

        """
        return "\n".join("".join(line) for line in self.display_map[::-1])

    def __repr__(self):
        nnodes = 0
        if self.node_index_map:
            nnodes = len(self.node_index_map)
        return f"<XYMap(Z={self.Z}), {self.max_X + 1}x{self.max_Y + 1}, {nnodes} nodes>"

    def log(self, msg):
        if self.xyzgrid:
            self.xyzgrid.log(msg)
        else:
            logger.log_info(msg)

    def reload(self, map_module_or_dict=None):
        """
        (Re)Load a map.

        Args:
            map_module_or_dict (str, module or dict, optional): See description for the variable
                in the class' `__init__` function. If given, replace the already loaded
                map with a new one. If not given, the existing one given on class creation
                will be reloaded.
            parse (bool, optional): If set, auto-run `.parse()` on the newly loaded data.

        Notes:
            This will both (re)load the data and parse it into a new map structure, replacing any
            existing one. The valid mapstructure is:
            ::

                {
                    "map": <str>,
                    "zcoord": <int or str>, # optional
                    "legend": <dict>,       # optional
                    "prototypes": <dict>    # optional
                    "options": <dict>       # optional
                }

        """
        if not map_module_or_dict:
            map_module_or_dict = self.map_module_or_dict

        mapdata = {}
        if isinstance(map_module_or_dict, dict):
            # map-structure provided directly
            mapdata = map_module_or_dict
        else:
            # read from contents of module
            mod = mod_import(map_module_or_dict)
            mapdata_list = variable_from_module(mod, "XYMAP_DATA_LIST")
            if mapdata_list and self.Z:
                # use the stored Z value to figure out which map data we want
                mapping = {mapdata.get("zcoord") for mapdata in mapdata_list}
                mapdata = mapping.get(self.Z, {})

            if not mapdata:
                mapdata = variable_from_module(mod, "XYMAP_DATA")

        if not mapdata:
            raise MapError(
                "No valid XYMAP_DATA or XYMAP_DATA_LIST could be found from "
                f"{map_module_or_dict}."
            )

        # validate
        if any(key for key in mapdata if key not in MAP_DATA_KEYS):
            raise MapError(
                f"Mapdata has keys {list(mapdata)}, but only " f"keys {MAP_DATA_KEYS} are allowed."
            )

        for key in mapdata.get("legend", DEFAULT_LEGEND):
            if not key or len(key) > 1:
                if key not in self.legend_key_exceptions:
                    raise MapError(
                        f"Map-legend key '{key}' is invalid: All keys must "
                        "be exactly one character long. Use the node/link's "
                        "`.display_symbol` property to change how it is "
                        "displayed."
                    )
        if "map" not in mapdata or not mapdata["map"]:
            raise MapError("No map found. Add 'map' key to map-data dict.")
        for key, prototype in mapdata.get("prototypes", {}).items():
            if not (is_iter(key) and (2 <= len(key) <= 3)):
                raise MapError(
                    f"Prototype override key {key} is malformed: It must be a "
                    "coordinate (X, Y) for nodes or (X, Y, direction) for links; "
                    "where direction is a supported direction string ('n', 'ne', etc)."
                )

        # store/update result
        self.Z = mapdata.get("zcoord", self.Z)
        self.mapstring = mapdata["map"]
        self.prototypes = mapdata.get("prototypes", {})
        self.options = mapdata.get("options", {})

        # merge the custom legend onto the default legend to allow easily
        # overriding only parts of it
        self.legend = {**DEFAULT_LEGEND, **map_module_or_dict.get("legend", DEFAULT_LEGEND)}

        # initialize any prototypes on the legend entities
        for char, node_or_link_class in self.legend.items():
            prototype = node_or_link_class.prototype
            if not prototype or isinstance(prototype, dict):
                # nothing more to do
                continue
            # we need to load the prototype dict onto each for ease of access. Note that
            proto = protlib.search_prototype(
                prototype, require_single=True, no_db=_NO_DB_PROTOTYPES
            )[0]
            node_or_link_class.prototype = proto

    def parse(self):
        """
        Parses the numerical grid from the string. The first pass means parsing out
        all nodes. The linking-together of nodes is not happening until the second pass
        (the reason for this is that maps can also link to other maps, so all maps need
        to have gone through their first parsing-passes before they can be linked together).

        See the class docstring for details of how the grid should be defined.

        Notes:
            In this parsing, the 'xygrid' is the full range of chraracters read from
            the string. The `XYgrid` is used to denote the game-world coordinates
            (which doesn't include the links)

        """
        mapcorner_symbol = self.mapcorner_symbol
        # this allows for string-based [x][y] mapping with arbitrary objects
        xygrid = defaultdict(dict)
        # mapping nodes to real X,Y positions
        XYgrid = defaultdict(dict)
        # needed by pathfinder
        node_index_map = {}
        # used by transitions
        symbol_map = defaultdict(list)

        mapstring = self.mapstring
        if mapstring.count(mapcorner_symbol) < 2:
            raise MapParserError(
                f"The mapstring must have at least two '{mapcorner_symbol}' "
                "symbols marking the upper- and bottom-left corners of the "
                "grid area."
            )

        # find the the position (in the string as a whole) of the top-left corner-marker
        maplines = mapstring.split("\n")
        topleft_marker_x, topleft_marker_y = -1, -1
        for topleft_marker_y, line in enumerate(maplines):
            topleft_marker_x = line.find(mapcorner_symbol)
            if topleft_marker_x != -1:
                break
        if -1 in (topleft_marker_x, topleft_marker_y):
            raise MapParserError(f"No top-left corner-marker ({mapcorner_symbol}) found!")

        # find the position (in the string as a whole) of the bottom-left corner-marker
        # this is always in a stright line down from the first marker
        botleft_marker_x, botleft_marker_y = topleft_marker_x, -1
        for botleft_marker_y, line in enumerate(maplines[topleft_marker_y + 1 :]):
            if line.find(mapcorner_symbol) == topleft_marker_x:
                break
        if botleft_marker_y == -1:
            raise MapParserError(
                f"No bottom-left corner-marker ({mapcorner_symbol}) found! "
                "Make sure it lines up with the top-left corner-marker "
                f"(found at column {topleft_marker_x} of the string)."
            )
        # the actual coordinate is dy below the topleft marker so we need to shift
        botleft_marker_y += topleft_marker_y + 1

        # in-string_position of the top- and bottom-left grid corners (2 steps in from marker)
        # the bottom-left corner is also the origo (0,0) of the grid.
        topleft_y = topleft_marker_y + 2
        origo_x, origo_y = botleft_marker_x + 2, botleft_marker_y - 1

        # highest actually filled grid points
        max_x = 0
        max_y = 0
        max_X = 0
        max_Y = 0
        node_index = -1

        # first pass: read string-grid (left-right, bottom-up) and parse all grid points
        for iy, line in enumerate(reversed(maplines[topleft_y:origo_y])):
            even_iy = iy % 2 == 0
            for ix, char in enumerate(line[origo_x:]):
                # from now on, coordinates are on the xygrid.

                if char == self.empty_symbol:
                    continue

                # only set this if there's actually something on the line
                max_x, max_y = max(max_x, ix), max(max_y, iy)

                mapnode_or_link_class = self.legend.get(char)
                if not mapnode_or_link_class:
                    raise MapParserError(
                        f"Symbol '{char}' on XY=({ix / 2:g},{iy / 2:g}) " "is not found in LEGEND."
                    )
                if hasattr(mapnode_or_link_class, "node_index"):
                    # A mapnode. Mapnodes can only be placed on even grid positions, where
                    # there are integer X,Y coordinates defined.

                    if not (even_iy and ix % 2 == 0):
                        raise MapParserError(
                            f"Symbol '{char}' on XY=({ix / 2:g},{iy / 2:g}) marks a "
                            "MapNode but is located between integer (X,Y) positions (only "
                            "Links can be placed between coordinates)!"
                        )

                    # save the node to several different maps for different uses
                    # in both coordinate systems
                    iX, iY = ix // 2, iy // 2
                    max_X, max_Y = max(max_X, iX), max(max_Y, iY)
                    node_index += 1

                    xygrid[ix][iy] = XYgrid[iX][iY] = node_index_map[
                        node_index
                    ] = mapnode_or_link_class(
                        x=ix, y=iy, Z=self.Z, node_index=node_index, symbol=char, xymap=self
                    )

                else:
                    # we have a link at this xygrid position (this is ok everywhere)
                    xygrid[ix][iy] = mapnode_or_link_class(
                        x=ix, y=iy, Z=self.Z, symbol=char, xymap=self
                    )

                # store the symbol mapping for transition lookups
                symbol_map[char].append(xygrid[ix][iy])

        # store before building links
        self.max_x, self.max_y = max_x, max_y
        self.max_X, self.max_Y = max_X, max_Y
        self.xygrid = xygrid
        self.XYgrid = XYgrid
        self.node_index_map = node_index_map
        self.symbol_map = symbol_map

        # build all links
        for node in node_index_map.values():
            node.build_links()

        # build display map
        display_map = [[" "] * (max_x + 1) for _ in range(max_y + 1)]
        for ix, ydct in xygrid.items():
            for iy, node_or_link in ydct.items():
                display_map[iy][ix] = node_or_link.get_display_symbol()

        for node in node_index_map.values():
            # override node-prototypes, ignore if no prototype
            # is defined (some nodes should not be spawned)
            if node.prototype:
                node_coord = (node.X, node.Y)
                # load prototype from override, or use default
                try:
                    node.prototype = flatten_prototype(
                        self.prototypes.get(
                            node_coord, self.prototypes.get(("*", "*"), node.prototype)
                        ),
                        no_db=_NO_DB_PROTOTYPES,
                    )
                except Exception as err:
                    raise MapParserError(f"Room prototype malformed: {err}", node)
                # do the same for links (x, y, direction) coords
                for direction, maplink in node.first_links.items():
                    try:
                        maplink.prototype = flatten_prototype(
                            self.prototypes.get(
                                node_coord + (direction,),
                                self.prototypes.get(("*", "*", "*"), maplink.prototype),
                            ),
                            no_db=_NO_DB_PROTOTYPES,
                        )
                    except Exception as err:
                        raise MapParserError(f"Exit prototype malformed: {err}", maplink)

        # store
        self.display_map = display_map

    def _get_topology_around_coord(self, xy, dist=2):
        """
        Get all links and nodes up to a certain distance from an XY coordinate.

        Args:
            xy (tuple), the X,Y coordinate of the center point.
            dist (int): How many nodes away from center point to find paths for.

        Returns:
            tuple: A tuple of 5 elements `(xy_coords, xmin, xmax, ymin, ymax)`, where the
                first element is a list of xy-coordinates (on xygrid) for all linked nodes within
                range. This is meant to be used with the xygrid for extracting a subset
                for display purposes. The others are the minimum size of the rectangle
                surrounding the area containing `xy_coords`.

        Notes:
            This performs a depth-first pass down the the given dist.

        """

        def _scan_neighbors(
            start_node, points, dist=2, xmin=BIGVAL, ymin=BIGVAL, xmax=0, ymax=0, depth=0
        ):
            x0, y0 = start_node.x, start_node.y
            points.append((x0, y0))
            xmin, xmax = min(xmin, x0), max(xmax, x0)
            ymin, ymax = min(ymin, y0), max(ymax, y0)

            if depth < dist:
                # keep stepping
                for direction, end_node in start_node.links.items():
                    x, y = x0, y0
                    for link in start_node.xy_steps_to_node[direction]:
                        x, y = link.x, link.y
                        points.append((x, y))
                        xmin, xmax = min(xmin, x), max(xmax, x)
                        ymin, ymax = min(ymin, y), max(ymax, y)

                    points, xmin, xmax, ymin, ymax = _scan_neighbors(
                        end_node,
                        points,
                        dist=dist,
                        xmin=xmin,
                        ymin=ymin,
                        xmax=xmax,
                        ymax=ymax,
                        depth=depth + 1,
                    )

            return points, xmin, xmax, ymin, ymax

        center_node = self.get_node_from_coord(xy)
        points, xmin, xmax, ymin, ymax = _scan_neighbors(center_node, [], dist=dist)
        return list(set(points)), xmin, xmax, ymin, ymax

    def calculate_path_matrix(self, force=False):
        """
        Solve the pathfinding problem using Dijkstra's algorithm. This will try to
        load the solution from disk if possible.

        Args:
            force (bool, optional): If the cache should always be rebuilt.

        """
        if not force and self.pathfinder_baked_filename and isfile(self.pathfinder_baked_filename):
            # check if the solution for this grid was already solved previously.

            mapstr, dist_matrix, pathfinding_routes = "", None, None
            with open(self.pathfinder_baked_filename, "rb") as fil:
                try:
                    mapstr, dist_matrix, pathfinding_routes = pickle.load(fil)
                except Exception:
                    logger.log_trace()
            if (
                mapstr == self.mapstring
                and dist_matrix is not None
                and pathfinding_routes is not None
            ):
                # this is important - it means the map hasn't changed so
                # we can re-use the stored data!
                self.dist_matrix = dist_matrix
                self.pathfinding_routes = pathfinding_routes

        # build a matrix representing the map graph, with 0s as impassable areas

        nnodes = len(self.node_index_map)
        pathfinding_graph = zeros((nnodes, nnodes))
        for inode, node in self.node_index_map.items():
            pathfinding_graph[inode, :] = node.linkweights(nnodes)

        # create a sparse matrix to represent link relationships from each node
        pathfinding_matrix = csr_matrix(pathfinding_graph)

        # solve using Dijkstra's algorithm
        self.dist_matrix, self.pathfinding_routes = dijkstra(
            pathfinding_matrix,
            directed=True,
            return_predecessors=True,
            limit=self.max_pathfinding_length,
        )

        if self.pathfinder_baked_filename:
            # try to cache the results
            with open(self.pathfinder_baked_filename, "wb") as fil:
                pickle.dump(
                    (self.mapstring, self.dist_matrix, self.pathfinding_routes), fil, protocol=4
                )

    def spawn_nodes(self, xy=("*", "*")):
        """
        Convert the nodes of this XYMap into actual in-world rooms by spawning their
        related prototypes in the correct coordinate positions. This must be done *first*
        before spawning links (with `spawn_links` because exits require the target destination
        to exist. It's also possible to only spawn a subset of the map

        Args:
            xy (tuple, optional): An (X,Y) coordinate of node(s). `'*'` acts as a wildcard.

        Examples:
            - `xy=(1, 3) - spawn (1,3) coordinate only.
            - `xy=('*', 1) - spawn all nodes in the first row of the map only.
            - `xy=('*', '*')` - spawn all nodes

        Returns:
            list: A list of nodes that were spawned.

        """
        global _XYZROOMCLASS
        if not _XYZROOMCLASS:
            from evennia.contrib.grid.xyzgrid.xyzroom import XYZRoom as _XYZROOMCLASS
        x, y = xy
        wildcard = "*"
        spawned = []

        # find existing nodes, in case some rooms need to be removed
        map_coords = [
            (node.X, node.Y)
            for node in sorted(self.node_index_map.values(), key=lambda n: (n.Y, n.X))
        ]
        for existing_room in _XYZROOMCLASS.objects.filter_xyz(xyz=(x, y, self.Z)):
            roomX, roomY, _ = existing_room.xyz
            if (roomX, roomY) not in map_coords:
                self.log(f"  deleting room at {existing_room.xyz} (not found on map).")
                existing_room.delete()

        # (re)build nodes (will not build already existing rooms)
        for node in sorted(self.node_index_map.values(), key=lambda n: (n.Y, n.X)):
            if (x in (wildcard, node.X)) and (y in (wildcard, node.Y)):
                node.spawn()
                spawned.append(node)
        return spawned

    def spawn_links(self, xy=("*", "*"), nodes=None, directions=None):
        """
        Convert links of this XYMap into actual in-game exits by spawning their related
        prototypes. It's possible to only spawn a specic exit by specifying the node and

        Args:
            xy (tuple, optional): An (X,Y) coordinate of node(s). `'*'` acts as a wildcard.
            nodes (list, optional): If given, only consider links out of these nodes. This also
                affects `xy`, so that if there are no nodes of given coords in `nodes`, no
                links will be spawned at all.
            directions (list, optional): A list of cardinal directions ('n', 'ne' etc). If given,
                sync only the exit in the given directions (`xy` limits which links out of which
                nodes should be considered). If unset, there are no limits to directions.
        Examples:
            - `xy=(1, 3 )`, `direction='ne'` - sync only the north-eastern exit
                out of the (1, 3) node.

        """
        x, y = xy
        wildcard = "*"

        if not nodes:
            nodes = sorted(self.node_index_map.values(), key=lambda n: (n.Z, n.Y, n.X))

        for node in nodes:
            if (x in (wildcard, node.X)) and (y in (wildcard, node.Y)):
                node.spawn_links(directions=directions)

    def get_node_from_coord(self, xy):
        """
        Get a MapNode from a coordinate.

        Args:
            xy (tuple): X,Y coordinate on XYgrid.

        Returns:
            MapNode: The node found at the given coordinates. Returns
                `None` if there is no mapnode at the given coordinate.

        Raises:
            MapError: If trying to specify an iX,iY outside
                of the grid's maximum bounds.

        """
        if not self.XYgrid:
            self.parse()

        iX, iY = xy
        if not ((0 <= iX <= self.max_X) and (0 <= iY <= self.max_Y)):
            raise MapError(
                f"get_node_from_coord got coordinate {xy} which is "
                f"outside the grid size of (0,0) - ({self.max_X}, {self.max_Y})."
            )
        try:
            return self.XYgrid[iX][iY]
        except KeyError:
            return None

    def get_components_with_symbol(self, symbol):
        """
        Find all map components (nodes, links) with a given symbol in this map.

        Args:
            symbol (char): A single character-symbol to search for.

        Returns:
            list: A list of MapNodes and/or MapLinks found with the matching symbol.

        """
        return self.symbol_map.get(symbol, [])

    def get_shortest_path(self, start_xy, end_xy):
        """
        Get the shortest route between two points on the grid.

        Args:
            start_xy (tuple): A starting (X,Y) coordinate on the XYgrid (in-game coordinate) for
                where we start from.
            end_xy (tuple or MapNode): The end (X,Y) coordinate on the XYgrid (in-game coordinate)
                we want to find the shortest route to.

        Returns:
            tuple: Two lists, first containing the list of directions as strings (n, ne etc) and
            the second is a mixed list of MapNodes and all MapLinks in a sequence describing
            the full path including the start- and end-node.

        """
        startnode = self.get_node_from_coord(start_xy)
        endnode = self.get_node_from_coord(end_xy)

        if not (startnode and endnode):
            # no node at given coordinate. No path is possible.
            return [], []

        try:
            istartnode = startnode.node_index
            inextnode = endnode.node_index
        except AttributeError:
            raise MapError(
                f"Map.get_shortest_path received start/end nodes {startnode} and "
                f"{endnode}. They must both be MapNodes (not Links)"
            )

        if self.pathfinding_routes is None:
            self.calculate_path_matrix()

        pathfinding_routes = self.pathfinding_routes
        node_index_map = self.node_index_map

        path = [endnode]
        directions = []

        while pathfinding_routes[istartnode, inextnode] != -9999:
            # the -9999 is set by algorithm for unreachable nodes or if trying
            # to go a node we are already at (the start node in this case since
            # we are working backwards).
            inextnode = pathfinding_routes[istartnode, inextnode]
            nextnode = node_index_map[inextnode]
            shortest_route_to = nextnode.shortest_route_to_node[path[-1].node_index]

            directions.append(shortest_route_to[0])
            path.extend(shortest_route_to[1][::-1] + [nextnode])

        # we have the path - reverse to get the correct order
        directions = directions[::-1]
        path = path[::-1]

        return directions, path

    def get_visual_range(
        self,
        xy,
        dist=2,
        mode="nodes",
        character="@",
        target=None,
        target_path_style="|y{display_symbol}|n",
        max_size=None,
        indent=0,
        return_str=True,
    ):
        """
        Get a part of the grid centered on a specific point and extended a certain number
        of nodes or grid points in every direction.

        Args:
            xy (tuple): (X,Y) in-world coordinate location. If this is not the location
                of a node on the grid, the `character` or the empty-space symbol (by default
                an empty space) will be shown.
            dist (int, optional): Number of gridpoints distance to show. Which
                grid to use depends on the setting of `only_nodes`. Set to `None` to
                always show the entire grid.
            mode (str, optional): One of 'scan' or 'nodes'. In 'scan' mode, dist measure
                number of xy grid points in all directions and doesn't care about if visible
                nodes are reachable or not. If 'nodes', distance measure how many linked nodes
                away from the center coordinate to display.
            character (str, optional): Place this symbol at the `xy` position
                of the displayed map. The center node's symbol is shown if this is falsy.
            target (tuple, optional): A target XY coordinate to go to. The path to this
                (or the beginning of said path, if outside of visual range) will be
                marked according to `target_path_style`.
            target_path_style (str or callable, optional): This is use for marking the path
                found when `target` is given. If a string, it accepts a formatting marker
                `display_symbol` which will be filled with the `display_symbol` of each node/link
                the path passes through. This allows e.g. to color the path. If a callable, this
                will receive the MapNode or MapLink object for every step of the path and and
                must return the suitable string to display at the position of the node/link.
            max_size (tuple, optional): A max `(width, height)` to crop the displayed
                return to. Make both odd numbers to get a perfect center. Set either of
                the tuple values to `None` to make that coordinate unlimited. Set entire
                tuple to None let display-size able to grow up to full size of grid.
            indent (int, optional): How far to the right to indent the map area (only
                applies to `return_str=True`).
            return_str (bool, optional): Return result as an already formatted string
                or a 2D list.

        Returns:
            str or list: Depending on value of `return_str`. If a list,
                this is 2D grid of lines, [[str,str,str,...], [...]] where
                each element is a single character in the display grid. To
                extract a character at (ix,iy) coordinate from it, use
                indexing `outlist[iy][ix]` in that order.

        Notes:
            If outputting a list, the y-axis must first be reversed before printing since printing
            happens top-bottom and the y coordinate system goes bottom-up. This can be done simply
            with this before building the final string to send/print.

                printable_order_list = outlist[::-1]

            If mode='nodes', a `dist` of 2 will give the following result in a row of nodes:

               #-#-@----------#-#

            This display may thus visually grow much bigger than expected (both horizontally and
            vertically). consider setting `max_size` if wanting to restrict the display size. Also
            note that link 'weights' are *included* in this estimate, so if links have weights > 1,
            fewer nodes may be found for a given `dist`.

            If mode=`scan`, a dist of 2 on the above example would instead give

                #-@--

            This mode simply shows a cut-out subsection of the map you are on. The `dist` is
            measured on xygrid, so two steps per XY coordinate. It does not consider links or
            weights and may also show nodes not actually reachable at the moment:

                | |
                # @-#

        """
        iX, iY = xy
        # convert inputs to xygrid
        width, height = self.max_x + 1, self.max_y + 1
        ix, iy = max(0, min(iX * 2, width)), max(0, min(iY * 2, height))
        display_map = self.display_map
        xmin, xmax, ymin, ymax = 0, width - 1, 0, height - 1

        if dist is None:
            # show the entire grid
            gridmap = self.display_map
            ixc, iyc = ix, iy

        elif dist is None or dist <= 0 or not self.get_node_from_coord(xy):
            # There is no node at these coordinates. Show
            # nothing but ourselves or emptiness
            return character if character else self.empty_symbol

        elif mode == "nodes":
            # dist measures only full, reachable nodes.
            points, xmin, xmax, ymin, ymax = self._get_topology_around_coord(xy, dist=dist)

            ixc, iyc = ix - xmin, iy - ymin
            # note - override width/height here since our grid is
            # now different from the original for future cropping
            width, height = xmax - xmin + 1, ymax - ymin + 1
            gridmap = [[" "] * width for _ in range(height)]
            for ix0, iy0 in points:
                gridmap[iy0 - ymin][ix0 - xmin] = display_map[iy0][ix0]

        elif mode == "scan":
            # scan-mode - dist measures individual grid points

            xmin, xmax = max(0, ix - dist), min(width, ix + dist + 1)
            ymin, ymax = max(0, iy - dist), min(height, iy + dist + 1)
            ixc, iyc = ix - xmin, iy - ymin
            gridmap = [line[xmin:xmax] for line in display_map[ymin:ymax]]

        else:
            raise MapError(
                f"Map.get_visual_range 'mode' was '{mode}' "
                "- it must be either 'scan' or 'nodes'."
            )
        if character:
            gridmap[iyc][ixc] = character  # correct indexing; it's a list of lines

        if target:
            # stylize path to target

            def _default_callable(node):
                return target_path_style.format(display_symbol=node.get_display_symbol())

            if callable(target_path_style):
                _target_path_style = target_path_style
            else:
                _target_path_style = _default_callable

            _, path = self.get_shortest_path(xy, target)

            maxstep = dist if mode == "nodes" else dist / 2
            nsteps = 0
            for node_or_link in path[1:]:
                if hasattr(node_or_link, "node_index"):
                    nsteps += 1
                if nsteps > maxstep:
                    break
                # don't decorate current (character?) location
                ix, iy = node_or_link.x, node_or_link.y
                if xmin <= ix <= xmax and ymin <= iy <= ymax:
                    gridmap[iy - ymin][ix - xmin] = _target_path_style(node_or_link)

        if max_size:
            # crop grid to make sure it doesn't grow too far
            max_x, max_y = max_size
            max_x = self.max_x if max_x is None else max_x
            max_y = self.max_y if max_y is None else max_y
            xmin, xmax = max(0, ixc - max_x // 2), min(width, ixc + max_x // 2 + 1)
            ymin, ymax = max(0, iyc - max_y // 2), min(height, iyc + max_y // 2 + 1)
            gridmap = [line[xmin:xmax] for line in gridmap[ymin:ymax]]

        if return_str:
            # we must flip the y-axis before returning the string
            indent = indent * " "
            return indent + f"\n{indent}".join("".join(line) for line in gridmap[::-1])
        else:
            return gridmap



"""
This module stores session-level commands.
"""
from evennia.commands.cmdset import CmdSet
from evennia.commands.default import account


class SessionCmdSet(CmdSet):
    """
    Sets up the unlogged cmdset.
    """

    key = "DefaultSession"
    priority = -20

    def at_cmdset_creation(self):
        "Populate the cmdset"
        self.add(account.CmdSessions())



"""
Basic Map - helpme 2022

This adds an ascii `map` to a given room which can be viewed with the `map` command.
You can easily alter it to add special characters, room colors etc. The map shown is
dynamically generated on use, and supports all compass directions and up/down. Other
directions are ignored.

If you don't expect the map to be updated frequently, you could choose to save the
calculated map as a .ndb value on the room and render that instead of running mapping
calculations anew each time.

An example map:
```
       |
     -[-]-
       |
       |
-[-]--[-]--[-]--[-]
  |    |    |    |
       |    |    |
     -[-]--[-]  [-]
       | \/ |    |
     \ | /\ |
     -[-]--[-]
```

Installation:

Adding the `MapDisplayCmdSet` to the default character cmdset will add the `map` command.

Specifically, in `mygame/commands/default_cmdsets.py`:

```
...
from evennia.contrib.grid.ingame_map_display import MapDisplayCmdSet  # <---

class CharacterCmdset(default_cmds.Character_CmdSet):
    ...
    def at_cmdset_creation(self):
        ...
        self.add(MapDisplayCmdSet)  # <---

```

Then `reload` to make the new commands available.

Additional Settings:

In order to change your default map size, you can add to `mygame/server/settings.py`:

BASIC_MAP_SIZE = 5

This changes the default map width/height. 2-5 for most clients is sensible.

If you don't want the player to be able to specify the size of the map, ignore any
arguments passed into the Map command.
"""
import time

from django.conf import settings

from evennia import CmdSet
from evennia.commands.default.muxcommand import MuxCommand

_BASIC_MAP_SIZE = settings.BASIC_MAP_SIZE if hasattr(settings, "BASIC_MAP_SIZE") else 2
_MAX_MAP_SIZE = settings.BASIC_MAP_SIZE if hasattr(settings, "MAX_MAP_SIZE") else 10

# _COMPASS_DIRECTIONS specifies which way to move the pointer on the x/y axes and what characters to use to depict the exits on the map.
_COMPASS_DIRECTIONS = {
    "north": (0, -3, " | "),
    "south": (0, 3, " | "),
    "east": (3, 0, "-"),
    "west": (-3, 0, "-"),
    "northeast": (3, -3, "/"),
    "northwest": (-3, -3, "\\"),
    "southeast": (3, 3, "\\"),
    "southwest": (-3, 3, "/"),
    "up": (0, 0, "^"),
    "down": (0, 0, "v"),
}


class Map(object):
    def __init__(self, caller, size=_BASIC_MAP_SIZE, location=None):
        """
        Initializes the map.

        Args:
            caller (object): Any object, though generally a puppeted character.
            size (int): The seed size of the map, which will be multiplied to get the final grid size.
            location (object): The location at the map's center (will default to caller.location if none provided).
        """
        self.start_time = time.time()
        self.caller = caller
        self.max_width = int(size * 2 + 1) * 5  # This must be an odd number
        self.max_length = int(size * 2 + 1) * 3  # This must be an odd number
        self.has_mapped = {}
        self.curX = None
        self.curY = None
        self.size = size
        self.location = location or caller.location

    def create_grid(self):
        """
        Create the empty grid for the map based on the configured size

        Returns:
            list: The created grid, a list of lists.
        """
        board = []
        for row in range(self.max_length):
            board.append([])
            for column in range(int(self.max_width / 5)):
                board[row].extend([" ", "   ", " "])
        return board

    def exit_name_as_ordinal(self, ex):
        """
        Get the exit name as a compass direction if possible

        Args:
            ex (Exit): The current exit being mapped.
        Returns:
            string: The exit name as a compass direction or an empty string.
        """
        exit_name = ex.name
        if exit_name not in _COMPASS_DIRECTIONS:
            compass_aliases = [
                direction in ex.aliases.all() for direction in _COMPASS_DIRECTIONS.keys()
            ]
            if compass_aliases[0]:
                exit_name = compass_aliases[0]
            if exit_name not in _COMPASS_DIRECTIONS:
                return ""
        return exit_name

    def update_pos(self, room, exit_name):
        """
        Update the position pointer.

        Args:
            room (Room): The current location.
            exit_name (str): The name of the exit to to use in this room. This must
                be a valid compass direction, or an error will be raised.
        Raises:
            KeyError: If providing a non-compass exit name.
        """
        # Update the pointer
        self.curX, self.curY = self.has_mapped[room][0], self.has_mapped[room][1]

        # Move the pointer depending on which direction the exit lies
        # exit_name has already been validated as an ordinal direction at this point
        self.curY += _COMPASS_DIRECTIONS[exit_name][0]
        self.curX += _COMPASS_DIRECTIONS[exit_name][1]

    def has_drawn(self, room):
        """
        Checks if the given room has already been drawn or not

        Args:
            room (Room): Room to check.
        Returns:
            bool: Whether or not the room has been drawn.
        """
        return True if room in self.has_mapped.keys() else False

    def draw_room_on_map(self, room, max_distance):
        """
        Draw the room and its exits on the map recursively

        Args:
            room (Room): The room to draw out.
            max_distance (int): How extensive the map is.
        """
        self.draw(room)
        self.draw_exits(room)

        if max_distance == 0:
            return

        # Check if the caller has access to the room in question. If not, don't draw it.
        # Additionally, if the name of the exit is not ordinal but an alias of it is, use that.
        for ex in [x for x in room.exits if x.access(self.caller, "traverse")]:
            ex_name = self.exit_name_as_ordinal(ex)
            if not ex_name or ex_name in ["up", "down"]:
                continue
            if self.has_drawn(ex.destination):
                continue

            self.update_pos(room, ex_name.lower())
            self.draw_room_on_map(ex.destination, max_distance - 1)

    def draw_exits(self, room):
        """
        Draw a given room's exit paths

        Args:
            room (Room): The room to draw exits of.
        """
        x, y = self.curX, self.curY
        for ex in room.exits:
            ex_name = self.exit_name_as_ordinal(ex)
            if not ex_name:
                continue

            ex_character = _COMPASS_DIRECTIONS[ex_name][2]
            delta_x = int(_COMPASS_DIRECTIONS[ex_name][1] / 3)
            delta_y = int(_COMPASS_DIRECTIONS[ex_name][0] / 3)

            # Make modifications if the exit has BOTH up and down exits
            if ex_name == "up":
                if "v" in self.grid[x][y]:
                    self.render_room(room, x, y, p1="^", p2="v")
                else:
                    self.render_room(room, x, y, here="^")
            elif ex_name == "down":
                if "^" in self.grid[x][y]:
                    self.render_room(room, x, y, p1="^", p2="v")
                else:
                    self.render_room(room, x, y, here="v")
            else:
                self.grid[x + delta_x][y + delta_y] = ex_character

    def draw(self, room):
        """
        Draw the map starting from a given room and add it to the cache of mapped rooms

        Args:
            room (Room): The room to render.
        """
        # draw initial caller location on map first!
        if room == self.location:
            self.start_loc_on_grid(room)
            self.has_mapped[room] = [self.curX, self.curY]
        else:
            # map all other rooms
            self.has_mapped[room] = [self.curX, self.curY]
            self.render_room(room, self.curX, self.curY)

    def render_room(self, room, x, y, p1="[", p2="]", here=None):
        """
        Draw a given room with ascii characters

        Args:
            room (Room): The room to render.
            x (int): The x-value of the room on the grid (horizontally, east/west).
            y (int): The y-value of the room on the grid (vertically, north/south).
            p1 (str): The first character of the 3-character room depiction.
            p2 (str): The last character of the 3-character room depiction.
            here (str): Defaults to none, a special character depicting the room.
        """
        # Note: This is where you would set colors, symbols etc.
        # Render the room
        you = list("[ ]")

        you[0] = f"{p1}|n"
        you[1] = f"{here if here else you[1]}"
        if room == self.caller.location:
            you[1] = "|[x|co|n"  # Highlight the location you are currently in
        you[2] = f"{p2}|n"

        self.grid[x][y] = "".join(you)

    def start_loc_on_grid(self, room):
        """
        Set the starting location on the grid based on the maximum width and length

        Args:
            room (Room): The room to begin with.
        """
        x = int((self.max_width * 0.6 - 1) / 2)
        y = int((self.max_length - 1) / 2)

        self.render_room(room, x, y)
        self.curX, self.curY = x, y

    def show_map(self, debug=False):
        """
        Create and show the map, piecing it all together in the end

        Args:
            debug (bool): Whether or not to return the time taken to build the map.
        """
        map_string = ""
        self.grid = self.create_grid()
        self.draw_room_on_map(self.location, self.size)

        for row in self.grid:
            map_row = "".join(row)
            if map_row.strip() != "":
                map_string += f"{map_row}\n"

        elapsed = time.time() - self.start_time
        if debug:
            map_string += f"\nTook {elapsed}ms to render the map.\n"

        return "%s" % map_string


class CmdMap(MuxCommand):
    """
    Check the local map around you.

    Usage: map (optional size)
    """

    key = "map"

    def func(self):
        size = _BASIC_MAP_SIZE
        max_size = _MAX_MAP_SIZE
        if self.args.isnumeric():
            size = min(max_size, int(self.args))

        # You can run show_map(debug=True) to see how long it takes.
        map_here = Map(self.caller, size=size).show_map()
        self.caller.msg((map_here, {"type": "map"}))


# CmdSet for easily install all commands
class MapDisplayCmdSet(CmdSet):
    """
    The map command.
    """

    def at_cmdset_creation(self):
        self.add(CmdMap)



"""
Account

The Account represents the game "account" and each login has only one
Account object. An Account is what chats on default channels but has no
other in-game-world existence. Rather the Account puppets Objects (such
as Characters) in order to actually participate in the game world.


Guest

Guest accounts are simple low-level accounts that are created/deleted
on the fly and allows users to test the game without the commitment
of a full registration. Guest accounts are deactivated by default; to
activate them, add the following line to your settings file:

    GUEST_ENABLED = True

You will also need to modify the connection screen to reflect the
possibility to connect with a guest account. The setting file accepts
several more options for customizing the Guest account system.

"""

from evennia.accounts.accounts import DefaultAccount, DefaultGuest


class Account(DefaultAccount):
    """
    This class describes the actual OOC account (i.e. the user connecting
    to the MUD). It does NOT have visual appearance in the game world (that
    is handled by the character which is connected to this). Comm channels
    are attended/joined using this object.

    It can be useful e.g. for storing configuration options for your game, but
    should generally not hold any character-related info (that's best handled
    on the character level).

    Can be set using BASE_ACCOUNT_TYPECLASS.


    * available properties

     key (string) - name of account
     name (string)- wrapper for user.username
     aliases (list of strings) - aliases to the object. Will be saved to database as AliasDB entries but returned as strings.
     dbref (int, read-only) - unique #id-number. Also "id" can be used.
     date_created (string) - time stamp of object creation
     permissions (list of strings) - list of permission strings

     user (User, read-only) - django User authorization object
     obj (Object) - game object controlled by account. 'character' can also be used.
     sessions (list of Sessions) - sessions connected to this account
     is_superuser (bool, read-only) - if the connected user is a superuser

    * Handlers

     locks - lock-handler: use locks.add() to add new lock strings
     db - attribute-handler: store/retrieve database attributes on this self.db.myattr=val, val=self.db.myattr
     ndb - non-persistent attribute handler: same as db but does not create a database entry when storing data
     scripts - script-handler. Add new scripts to object with scripts.add()
     cmdset - cmdset-handler. Use cmdset.add() to add new cmdsets to object
     nicks - nick-handler. New nicks with nicks.add().

    * Helper methods

     msg(text=None, **kwargs)
     execute_cmd(raw_string, session=None)
     search(ostring, global_search=False, attribute_name=None, use_nicks=False, location=None, ignore_errors=False, account=False)
     is_typeclass(typeclass, exact=False)
     swap_typeclass(new_typeclass, clean_attributes=False, no_default=True)
     access(accessing_obj, access_type='read', default=False)
     check_permstring(permstring)

    * Hook methods (when re-implementation, remember methods need to have self as first arg)

     basetype_setup()
     at_account_creation()

     - note that the following hooks are also found on Objects and are
       usually handled on the character level:

     at_init()
     at_cmdset_get(**kwargs)
     at_first_login()
     at_post_login(session=None)
     at_disconnect()
     at_message_receive()
     at_message_send()
     at_server_reload()
     at_server_shutdown()

    """

    pass


class Guest(DefaultGuest):
    """
    This class is used for guest logins. Unlike Accounts, Guests and their
    characters are deleted after disconnection.
    """

    pass



"""
Unit tests for the utilities of the evennia.utils.utils module.

TODO: Not nearly all utilities are covered yet.

"""

import os.path
import random
from datetime import datetime, timedelta

import mock
from django.test import TestCase
from parameterized import parameterized
from twisted.internet import task

from evennia.utils import utils
from evennia.utils.ansi import ANSIString
from evennia.utils.test_resources import BaseEvenniaTest


class TestIsIter(TestCase):
    def test_is_iter(self):
        self.assertEqual(True, utils.is_iter([1, 2, 3, 4]))
        self.assertEqual(False, utils.is_iter("This is not an iterable"))


class TestCrop(TestCase):
    def test_crop(self):
        # No text, return no text
        self.assertEqual("", utils.crop("", width=10, suffix="[...]"))
        # Input length equal to max width, no crop
        self.assertEqual("0123456789", utils.crop("0123456789", width=10, suffix="[...]"))
        # Input length greater than max width, crop (suffix included in width)
        self.assertEqual("0123[...]", utils.crop("0123456789", width=9, suffix="[...]"))
        # Input length less than desired width, no crop
        self.assertEqual("0123", utils.crop("0123", width=9, suffix="[...]"))
        # Width too small or equal to width of suffix
        self.assertEqual("012", utils.crop("0123", width=3, suffix="[...]"))
        self.assertEqual("01234", utils.crop("0123456", width=5, suffix="[...]"))


class TestDedent(TestCase):
    def test_dedent(self):
        # Empty string, return empty string
        self.assertEqual("", utils.dedent(""))
        # No leading whitespace
        self.assertEqual("TestDedent", utils.dedent("TestDedent"))
        # Leading whitespace, single line
        self.assertEqual("TestDedent", utils.dedent("   TestDedent"))
        # Leading whitespace, multi line
        input_string = "  hello\n  world"
        expected_string = "hello\nworld"
        self.assertEqual(expected_string, utils.dedent(input_string))


class TestListToString(TestCase):
    """
    Default function header from time.py:
    list_to_string(inlist, sep=",", endsep=", and", addquote=False)

    Examples:
     with defaults:
        [1,2,3] -> '1, 2, and 3'
     with endsep==',':
        [1,2,3] -> '1, 2, 3'
     with sep==';' and endsep==';':
        [1,2,3] -> '1; 2; 3'
     with sep=='or':
        [1,2,3] -> '1 or 2, and 3'
     with endsep=='and':
        [1,2,3] -> '1, 2 and 3'
     with endsep=='; and':
        [1,2,3] -> '1, 2; and 3'
     with endsep=='':
        [1,2,3] -> '1, 2 3'
     with addquote and endsep="and"
        [1,2,3] -> '"1", "2" and "3"'
    """

    def test_list_to_string(self):
        self.assertEqual("1, 2, and 3", utils.list_to_string([1, 2, 3]))
        self.assertEqual("1, 2, 3", utils.list_to_string([1, 2, 3], endsep=","))
        self.assertEqual("1, 2 and 3", utils.list_to_string([1, 2, 3], endsep="and"))
        self.assertEqual("1, 2 3", utils.list_to_string([1, 2, 3], endsep=""))
        self.assertEqual("1; 2; 3", utils.list_to_string([1, 2, 3], sep=";", endsep=";"))
        self.assertEqual("1 or 2, and 3", utils.list_to_string([1, 2, 3], sep="or"))
        self.assertEqual("1, 2; and 3", utils.list_to_string([1, 2, 3], endsep="; and"))
        self.assertEqual(
            '"1", "2", "3"', utils.list_to_string([1, 2, 3], endsep=",", addquote=True)
        )
        self.assertEqual(
            '"1", "2" and "3"', utils.list_to_string([1, 2, 3], endsep="and", addquote=True)
        )
        self.assertEqual("1 and 2", utils.list_to_string([1, 2]))
        self.assertEqual("1, 2", utils.list_to_string([1, 2], endsep=","))


class TestMLen(TestCase):
    """
    Verifies that m_len behaves like len in all situations except those
    where MXP may be involved.
    """

    def test_non_mxp_string(self):
        self.assertEqual(utils.m_len("Test_string"), 11)

    def test_mxp_string(self):
        self.assertEqual(utils.m_len("|lclook|ltat|le"), 2)

    def test_mxp_ansi_string(self):
        self.assertEqual(utils.m_len(ANSIString("|lcl|gook|ltat|le|n")), 2)

    def test_non_mxp_ansi_string(self):
        self.assertEqual(utils.m_len(ANSIString("|gHello|n")), 5)

    def test_list(self):
        self.assertEqual(utils.m_len([None, None]), 2)

    def test_dict(self):
        self.assertEqual(utils.m_len({"hello": True, "Goodbye": False}), 2)


class TestDisplayLen(TestCase):
    """
    Verifies that display_len behaves like m_len in all situations except those
    where asian characters are involved.
    """

    def test_non_mxp_string(self):
        self.assertEqual(utils.display_len("Test_string"), 11)

    def test_mxp_string(self):
        self.assertEqual(utils.display_len("|lclook|ltat|le"), 2)

    def test_mxp_ansi_string(self):
        self.assertEqual(utils.display_len(ANSIString("|lcl|gook|ltat|le|n")), 2)

    def test_non_mxp_ansi_string(self):
        self.assertEqual(utils.display_len(ANSIString("|gHello|n")), 5)

    def test_list(self):
        self.assertEqual(utils.display_len([None, None]), 2)

    def test_dict(self):
        self.assertEqual(utils.display_len({"hello": True, "Goodbye": False}), 2)

    def test_east_asian(self):
        self.assertEqual(utils.display_len("서서서"), 6)


class TestANSIString(TestCase):
    """
    Verifies that ANSIString's string-API works as intended.
    """

    def setUp(self):
        self.example_raw = "|relectric |cboogaloo|n"
        self.example_ansi = ANSIString(self.example_raw)
        self.example_str = "electric boogaloo"
        self.example_output = "\x1b[1m\x1b[31melectric \x1b[1m\x1b[36mboogaloo\x1b[0m"

    def test_length(self):
        self.assertEqual(len(self.example_ansi), 17)

    def test_clean(self):
        self.assertEqual(self.example_ansi.clean(), self.example_str)

    def test_raw(self):
        self.assertEqual(self.example_ansi.raw(), self.example_output)

    def test_format(self):
        self.assertEqual(f"{self.example_ansi:0<20}", self.example_output + "000")


class TestTimeformat(TestCase):
    """
    Default function header from time.py:
    time_format(seconds, style=0)

    """

    def test_style_0(self):
        """Test the style 0 of time_format."""
        self.assertEqual(utils.time_format(0, 0), "00:00")
        self.assertEqual(utils.time_format(28, 0), "00:00")
        self.assertEqual(utils.time_format(92, 0), "00:01")
        self.assertEqual(utils.time_format(300, 0), "00:05")
        self.assertEqual(utils.time_format(660, 0), "00:11")
        self.assertEqual(utils.time_format(3600, 0), "01:00")
        self.assertEqual(utils.time_format(3725, 0), "01:02")
        self.assertEqual(utils.time_format(86350, 0), "23:59")
        self.assertEqual(utils.time_format(86800, 0), "1d 00:06")
        self.assertEqual(utils.time_format(130800, 0), "1d 12:20")
        self.assertEqual(utils.time_format(530800, 0), "6d 03:26")

    def test_style_1(self):
        """Test the style 1 of time_format."""
        self.assertEqual(utils.time_format(0, 1), "0s")
        self.assertEqual(utils.time_format(28, 1), "28s")
        self.assertEqual(utils.time_format(92, 1), "1m")
        self.assertEqual(utils.time_format(300, 1), "5m")
        self.assertEqual(utils.time_format(660, 1), "11m")
        self.assertEqual(utils.time_format(3600, 1), "1h")
        self.assertEqual(utils.time_format(3725, 1), "1h")
        self.assertEqual(utils.time_format(86350, 1), "23h")
        self.assertEqual(utils.time_format(86800, 1), "1d")
        self.assertEqual(utils.time_format(130800, 1), "1d")
        self.assertEqual(utils.time_format(530800, 1), "6d")

    def test_style_2(self):
        """Test the style 2 of time_format."""
        self.assertEqual(utils.time_format(0, 2), "0 minutes")
        self.assertEqual(utils.time_format(28, 2), "0 minutes")
        self.assertEqual(utils.time_format(92, 2), "1 minute")
        self.assertEqual(utils.time_format(300, 2), "5 minutes")
        self.assertEqual(utils.time_format(660, 2), "11 minutes")
        self.assertEqual(utils.time_format(3600, 2), "1 hour, 0 minutes")
        self.assertEqual(utils.time_format(3725, 2), "1 hour, 2 minutes")
        self.assertEqual(utils.time_format(86350, 2), "23 hours, 59 minutes")
        self.assertEqual(utils.time_format(86800, 2), "1 day, 0 hours, 6 minutes")
        self.assertEqual(utils.time_format(130800, 2), "1 day, 12 hours, 20 minutes")
        self.assertEqual(utils.time_format(530800, 2), "6 days, 3 hours, 26 minutes")

    def test_style_3(self):
        """Test the style 3 of time_format."""
        self.assertEqual(utils.time_format(0, 3), "")
        self.assertEqual(utils.time_format(28, 3), "28 seconds")
        self.assertEqual(utils.time_format(92, 3), "1 minute 32 seconds")
        self.assertEqual(utils.time_format(300, 3), "5 minutes 0 seconds")
        self.assertEqual(utils.time_format(660, 3), "11 minutes 0 seconds")
        self.assertEqual(utils.time_format(3600, 3), "1 hour, 0 minutes")
        self.assertEqual(utils.time_format(3725, 3), "1 hour, 2 minutes 5 seconds")
        self.assertEqual(utils.time_format(86350, 3), "23 hours, 59 minutes 10 seconds")
        self.assertEqual(utils.time_format(86800, 3), "1 day, 0 hours, 6 minutes 40 seconds")
        self.assertEqual(utils.time_format(130800, 3), "1 day, 12 hours, 20 minutes 0 seconds")
        self.assertEqual(utils.time_format(530800, 3), "6 days, 3 hours, 26 minutes 40 seconds")

    def test_style_4(self):
        """Test the style 4 of time_format."""
        self.assertEqual(utils.time_format(0, 4), "0 seconds")
        self.assertEqual(utils.time_format(28, 4), "28 seconds")
        self.assertEqual(utils.time_format(92, 4), "a minute")
        self.assertEqual(utils.time_format(300, 4), "5 minutes")
        self.assertEqual(utils.time_format(660, 4), "11 minutes")
        self.assertEqual(utils.time_format(3600, 4), "an hour")
        self.assertEqual(utils.time_format(3725, 4), "an hour")
        self.assertEqual(utils.time_format(86350, 4), "23 hours")
        self.assertEqual(utils.time_format(86800, 4), "a day")
        self.assertEqual(utils.time_format(130800, 4), "a day")
        self.assertEqual(utils.time_format(530800, 4), "6 days")
        self.assertEqual(utils.time_format(3030800, 4), "a month")
        self.assertEqual(utils.time_format(7030800, 4), "2 months")
        self.assertEqual(utils.time_format(40030800, 4), "a year")
        self.assertEqual(utils.time_format(90030800, 4), "2 years")

    def test_unknown_format(self):
        """Test that unknown formats raise exceptions."""
        self.assertRaises(ValueError, utils.time_format, 0, 5)
        self.assertRaises(ValueError, utils.time_format, 0, "u")


@mock.patch(
    "evennia.utils.utils.timezone.now",
    new=mock.MagicMock(return_value=datetime(2019, 8, 28, 21, 56)),
)
class TestDateTimeFormat(TestCase):
    def test_datetimes(self):
        dtobj = datetime(2017, 7, 26, 22, 54)
        self.assertEqual(utils.datetime_format(dtobj), "Jul 26, 2017")
        dtobj = datetime(2019, 7, 26, 22, 54)
        self.assertEqual(utils.datetime_format(dtobj), "Jul 26")
        dtobj = datetime(2019, 8, 28, 19, 54)
        self.assertEqual(utils.datetime_format(dtobj), "19:54")
        dtobj = datetime(2019, 8, 28, 21, 32)
        self.assertEqual(utils.datetime_format(dtobj), "21:32:00")


class TestImportFunctions(TestCase):
    def _path_to_file(self, filename):
        testdir = os.path.dirname(os.path.abspath(__file__))
        return os.path.join(testdir, filename)

    def test_mod_import(self):
        loaded_mod = utils.mod_import("evennia.utils.ansi")
        self.assertIsNotNone(loaded_mod)

    def test_mod_import_invalid(self):
        loaded_mod = utils.mod_import("evennia.utils.invalid_module")
        self.assertIsNone(loaded_mod)

    def test_mod_import_from_path(self):
        test_path = self._path_to_file("test_eveditor.py")
        loaded_mod = utils.mod_import_from_path(test_path)
        self.assertIsNotNone(loaded_mod)

    def test_mod_import_from_path_invalid(self):
        test_path = self._path_to_file("invalid_filename.py")
        loaded_mod = utils.mod_import_from_path(test_path)
        self.assertIsNone(loaded_mod)


class LatinifyTest(TestCase):
    def setUp(self):
        super().setUp()

        self.example_str = "It naïvely says, “plugh.”"
        self.expected_output = 'It naively says, "plugh."'

    def test_plain_string(self):
        result = utils.latinify(self.example_str)
        self.assertEqual(result, self.expected_output)

    def test_byte_string(self):
        byte_str = utils.to_bytes(self.example_str)
        result = utils.latinify(byte_str)
        self.assertEqual(result, self.expected_output)


class TestFormatGrid(TestCase):
    maxDiff = None

    def setUp(self):
        # make the random only semi-random with a fixed seed
        random.seed(1)

    def tearDown(self):
        # restore normal randomness
        random.seed(None)

    def _generate_elements(self, basewidth, variation, amount):
        return [
            "X" * max(1, basewidth + int(random.randint(-variation, variation)))
            for _ in range(amount)
        ]

    def test_even_grid(self):
        """Grid with small variations"""
        elements = self._generate_elements(3, 1, 30)
        rows = utils.format_grid(elements, width=78)
        self.assertEqual(len(rows), 4)
        self.assertTrue(all(len(row) == 78 for row in rows))

    def test_disparate_grid(self):
        """Grid with big variations"""
        elements = self._generate_elements(3, 15, 30)
        rows = utils.format_grid(elements, width=82, sep="  ")
        self.assertEqual(len(rows), 8)
        self.assertTrue(all(len(row) == 82 for row in rows))

    def test_huge_grid(self):
        """Grid with very long strings"""
        elements = self._generate_elements(70, 20, 30)
        rows = utils.format_grid(elements, width=78)
        self.assertEqual(len(rows), 30)
        self.assertTrue(all(len(row) == 78 for row in rows))

    def test_overlap(self):
        """Grid with elements overlapping into the next slot"""
        elements = (
            "alias",
            "batchcode",
            "batchcommands",
            "cmdsets",
            "copy",
            "cpattr",
            "desc",
            "destroy",
            "dig",
            "examine",
            "find",
            "force",
            "lock",
        )
        rows = utils.format_grid(elements, width=78)
        self.assertEqual(len(rows), 3)
        for element in elements:
            self.assertTrue(element in "\n".join(rows), f"element {element} is missing.")

    def test_breakline(self):
        """Grid with line-long elements in middle"""
        elements = self._generate_elements(6, 4, 30)
        elements[10] = elements[20] = "-" * 78
        rows = utils.format_grid(elements, width=78)
        self.assertEqual(len(rows), 8)
        for element in elements:
            self.assertTrue(element in "\n".join(rows), f"element {element} is missing.")


class TestPercent(TestCase):
    """
    Test the utils.percentage function.
    """

    def test_ok_input(self):
        result = utils.percent(3, 0, 10)
        self.assertEqual(result, "30.0%")
        result = utils.percent(2.5, 0, 5, formatting=None)
        self.assertEqual(result, 50.0)

    def test_bad_input(self):
        """Gracefully handle weird input."""
        self.assertEqual(utils.percent(3, 10, 1), "0.0%")
        self.assertEqual(utils.percent(3, None, 1), "100.0%")
        self.assertEqual(utils.percent(1, 1, 1), "100.0%")
        self.assertEqual(utils.percent(3, 1, 1), "0.0%")
        self.assertEqual(utils.percent(3, 0, 1), "100.0%")
        self.assertEqual(utils.percent(-3, 0, 1), "0.0%")


class TestSafeConvert(TestCase):
    """
    Test evennia.utils.utils.safe_convert_to_types

    """

    @parameterized.expand(
        [
            (
                ("1", "2", 3, 4, "5"),
                {"a": 1, "b": "2", "c": 3},
                ((int, float, str, int), {"a": int, "b": float}),  # "
                (1, 2.0, "3", 4, "5"),
                {"a": 1, "b": 2.0, "c": 3},
            ),
            (
                ("1 + 2", "[1, 2, 3]", [3, 4, 5]),
                {"a": "3 + 4", "b": 5},
                (("py", "py", "py"), {"a": "py", "b": "py"}),
                (3, [1, 2, 3], [3, 4, 5]),
                {"a": 7, "b": 5},
            ),
        ]
    )
    def test_conversion(self, args, kwargs, converters, expected_args, expected_kwargs):
        """
        Test the converter with different inputs

        """
        result_args, result_kwargs = utils.safe_convert_to_types(
            converters, *args, raise_errors=True, **kwargs
        )
        self.assertEqual(expected_args, result_args)
        self.assertEqual(expected_kwargs, result_kwargs)

    def test_conversion__fail(self):
        """
        Test failing conversion

        """
        from evennia.utils.funcparser import ParsingError

        with self.assertRaises(ValueError):
            utils.safe_convert_to_types((int,), *("foo",), raise_errors=True)

        with self.assertRaises(ParsingError) as err:
            utils.safe_convert_to_types(("py", {}), *("foo",), raise_errors=True)


_TASK_HANDLER = None


def dummy_func(obj):
    """
    Used in TestDelay.

    A function that:
        can be serialized
        uses no memory references
        uses evennia objects
    """
    # get a reference of object
    from evennia.objects.models import ObjectDB

    obj = ObjectDB.objects.object_search(obj)
    obj = obj[0]
    # make changes to object
    obj.ndb.dummy_var = "dummy_func ran"
    return True


class TestDelay(BaseEvenniaTest):
    """
    Test utils.delay.
    """

    def setUp(self):
        super().setUp()
        # get a reference of TASK_HANDLER
        self.timedelay = 5
        global _TASK_HANDLER
        if _TASK_HANDLER is None:
            from evennia.scripts.taskhandler import TASK_HANDLER as _TASK_HANDLER
        _TASK_HANDLER.clock = task.Clock()
        self.char1.ndb.dummy_var = False

    def tearDown(self):
        super().tearDown()
        _TASK_HANDLER.clear()

    def test_call_early(self):
        # call a task early with call
        for pers in (True, False):
            t = utils.delay(self.timedelay, dummy_func, self.char1.dbref, persistent=pers)
            result = t.call()
            self.assertTrue(result)
            self.assertEqual(self.char1.ndb.dummy_var, "dummy_func ran")
            self.assertTrue(t.exists())
            self.assertTrue(t.active())
            self.char1.ndb.dummy_var = False

    def test_do_task(self):
        # call the task early with do_task
        for pers in (True, False):
            t = utils.delay(self.timedelay, dummy_func, self.char1.dbref, persistent=pers)
            # call the task early to test Task.call and TaskHandler.call_task
            result = t.do_task()
            self.assertTrue(result)
            self.assertEqual(self.char1.ndb.dummy_var, "dummy_func ran")
            self.assertFalse(t.exists())
            self.char1.ndb.dummy_var = False

    def test_deferred_call(self):
        # wait for deferred to call
        timedelay = self.timedelay
        for pers in (False, True):
            t = utils.delay(timedelay, dummy_func, self.char1.dbref, persistent=pers)
            self.assertTrue(t.active())
            _TASK_HANDLER.clock.advance(timedelay)  # make time pass
            self.assertEqual(self.char1.ndb.dummy_var, "dummy_func ran")
            self.assertFalse(t.exists())
            self.char1.ndb.dummy_var = False

    def test_short_deferred_call(self):
        # wait for deferred to call with a very short time
        timedelay = 0.1
        for pers in (False, True):
            t = utils.delay(timedelay, dummy_func, self.char1.dbref, persistent=pers)
            self.assertTrue(t.active())
            _TASK_HANDLER.clock.advance(timedelay)  # make time pass
            self.assertEqual(self.char1.ndb.dummy_var, "dummy_func ran")
            self.assertFalse(t.exists())
            self.char1.ndb.dummy_var = False

    def test_active(self):
        timedelay = self.timedelay
        t = utils.delay(timedelay, dummy_func, self.char1.dbref)
        self.assertTrue(_TASK_HANDLER.active(t.get_id()))
        self.assertTrue(t.active())
        _TASK_HANDLER.clock.advance(timedelay)  # make time pass
        self.assertFalse(_TASK_HANDLER.active(t.get_id()))
        self.assertFalse(t.active())

    def test_called(self):
        timedelay = self.timedelay
        t = utils.delay(timedelay, dummy_func, self.char1.dbref)
        self.assertFalse(t.called)
        _TASK_HANDLER.clock.advance(timedelay)  # make time pass
        self.assertTrue(t.called)

    def test_cancel(self):
        timedelay = self.timedelay
        for pers in (False, True):
            t = utils.delay(timedelay, dummy_func, self.char1.dbref, persistent=pers)
            self.assertTrue(t.active())
            success = t.cancel()
            self.assertFalse(t.active())
            self.assertTrue(success)
            self.assertTrue(t.exists())
            _TASK_HANDLER.clock.advance(timedelay)  # make time pass
            self.assertEqual(self.char1.ndb.dummy_var, False)

    def test_remove(self):
        timedelay = self.timedelay
        for pers in (False, True):
            t = utils.delay(timedelay, dummy_func, self.char1.dbref, persistent=pers)
            self.assertTrue(t.active())
            success = t.remove()
            self.assertTrue(success)
            self.assertFalse(t.active())
            self.assertFalse(t.exists())
            _TASK_HANDLER.clock.advance(timedelay)  # make time pass
            self.assertEqual(self.char1.ndb.dummy_var, False)

    def test_remove_canceled(self):
        # remove a canceled task
        timedelay = self.timedelay
        for pers in (False, True):
            t = utils.delay(timedelay, dummy_func, self.char1.dbref, persistent=pers)
            self.assertTrue(t.active())
            success = t.cancel()
            self.assertTrue(success)
            self.assertTrue(t.exists())
            self.assertFalse(t.active())
            success = t.remove()
            self.assertTrue(success)
            self.assertFalse(t.exists())
            _TASK_HANDLER.clock.advance(timedelay)  # make time pass
            self.assertEqual(self.char1.ndb.dummy_var, False)

    def test_pause_unpause(self):
        # remove a canceled task
        timedelay = self.timedelay
        for pers in (False, True):
            t = utils.delay(timedelay, dummy_func, self.char1.dbref, persistent=pers)
            self.assertTrue(t.active())
            t.pause()
            self.assertTrue(t.paused)
            t.unpause()
            self.assertFalse(t.paused)
            self.assertEqual(self.char1.ndb.dummy_var, False)
            t.pause()
            _TASK_HANDLER.clock.advance(timedelay)  # make time pass
            self.assertEqual(self.char1.ndb.dummy_var, False)
            t.unpause()
            self.assertEqual(self.char1.ndb.dummy_var, "dummy_func ran")
            self.char1.ndb.dummy_var = False

    def test_auto_stale_task_removal(self):
        # automated removal of stale tasks.
        timedelay = self.timedelay
        for pers in (False, True):
            t = utils.delay(timedelay, dummy_func, self.char1.dbref, persistent=pers)
            t.cancel()
            self.assertFalse(t.active())
            _TASK_HANDLER.clock.advance(timedelay)  # make time pass
            if pers:
                self.assertTrue(t.get_id() in _TASK_HANDLER.to_save)
            self.assertTrue(t.get_id() in _TASK_HANDLER.tasks)
            # Make task handler's now time, after the stale timeout
            _TASK_HANDLER._now = datetime.now() + timedelta(
                seconds=_TASK_HANDLER.stale_timeout + timedelay + 1
            )
            # add a task to test automatic removal
            t2 = utils.delay(timedelay, dummy_func, self.char1.dbref)
            if pers:
                self.assertFalse(t.get_id() in _TASK_HANDLER.to_save)
            self.assertFalse(t.get_id() in _TASK_HANDLER.tasks)
            self.assertEqual(self.char1.ndb.dummy_var, False)
            _TASK_HANDLER.clear()

    def test_manual_stale_task_removal(self):
        # manual removal of stale tasks.
        timedelay = self.timedelay
        for pers in (False, True):
            t = utils.delay(timedelay, dummy_func, self.char1.dbref, persistent=pers)
            t.cancel()
            self.assertFalse(t.active())
            _TASK_HANDLER.clock.advance(timedelay)  # make time pass
            if pers:
                self.assertTrue(t.get_id() in _TASK_HANDLER.to_save)
            self.assertTrue(t.get_id() in _TASK_HANDLER.tasks)
            # Make task handler's now time, after the stale timeout
            _TASK_HANDLER._now = datetime.now() + timedelta(
                seconds=_TASK_HANDLER.stale_timeout + timedelay + 1
            )
            _TASK_HANDLER.clean_stale_tasks()  # cleanup of stale tasks in in the save method
            if pers:
                self.assertFalse(t.get_id() in _TASK_HANDLER.to_save)
            self.assertFalse(t.get_id() in _TASK_HANDLER.tasks)
            self.assertEqual(self.char1.ndb.dummy_var, False)
            _TASK_HANDLER.clear()

    def test_disable_stale_removal(self):
        # manual removal of stale tasks.
        timedelay = self.timedelay
        _TASK_HANDLER.stale_timeout = 0
        for pers in (False, True):
            t = utils.delay(timedelay, dummy_func, self.char1.dbref, persistent=pers)
            t.cancel()
            self.assertFalse(t.active())
            _TASK_HANDLER.clock.advance(timedelay)  # make time pass
            if pers:
                self.assertTrue(t.get_id() in _TASK_HANDLER.to_save)
            self.assertTrue(t.get_id() in _TASK_HANDLER.tasks)
            # Make task handler's now time, after the stale timeout
            _TASK_HANDLER._now = datetime.now() + timedelta(
                seconds=_TASK_HANDLER.stale_timeout + timedelay + 1
            )
            t2 = utils.delay(timedelay, dummy_func, self.char1.dbref)
            if pers:
                self.assertTrue(t.get_id() in _TASK_HANDLER.to_save)
            self.assertTrue(t.get_id() in _TASK_HANDLER.tasks)
            self.assertEqual(self.char1.ndb.dummy_var, False)
            # manual removal should still work
            _TASK_HANDLER.clean_stale_tasks()  # cleanup of stale tasks in in the save method
            if pers:
                self.assertFalse(t.get_id() in _TASK_HANDLER.to_save)
            self.assertFalse(t.get_id() in _TASK_HANDLER.tasks)
            _TASK_HANDLER.clear()

    def test_server_restart(self):
        # emulate a server restart
        timedelay = self.timedelay
        utils.delay(timedelay, dummy_func, self.char1.dbref, persistent=True)
        _TASK_HANDLER.clear(False)  # remove all tasks from task handler, do not save this change.
        _TASK_HANDLER.clock.advance(timedelay)  # advance twisted reactor time past callback time
        self.assertEqual(self.char1.ndb.dummy_var, False)  # task has not run
        _TASK_HANDLER.load()  # load persistent tasks from database.
        _TASK_HANDLER.create_delays()  # create new deffered instances from persistent tasks
        _TASK_HANDLER.clock.advance(
            timedelay
        )  # Clock must advance to trigger, even if past timedelay
        self.assertEqual(self.char1.ndb.dummy_var, "dummy_func ran")


class TestIntConversions(TestCase):
    def test_int2str(self):
        self.assertEqual("three", utils.int2str(3))
        # special adjective conversion
        self.assertEqual("3rd", utils.int2str(3, adjective=True))
        # generic adjective conversion
        self.assertEqual("5th", utils.int2str(5, adjective=True))
        # No mapping return int as str
        self.assertEqual("15", utils.int2str(15))

    def test_str2int(self):
        # simple conversions
        self.assertEqual(5, utils.str2int("5"))

        # basic mapped numbers
        self.assertEqual(3, utils.str2int("three"))
        self.assertEqual(20, utils.str2int("twenty"))

        # multi-place numbers
        self.assertEqual(2345, utils.str2int("two thousand, three hundred and forty-five"))

        # ordinal numbers
        self.assertEqual(1, utils.str2int("1st"))
        self.assertEqual(1, utils.str2int("first"))
        self.assertEqual(4, utils.str2int("fourth"))
        # ordinal sound-change conversions
        self.assertEqual(5, utils.str2int("fifth"))
        self.assertEqual(20, utils.str2int("twentieth"))

        with self.assertRaises(ValueError):
            utils.str2int("not a number")


class TestJustify(TestCase):
    def test_justify_whitespace(self):
        result = utils.justify(" ", 1, align="l")
        self.assertEqual(" ", result)

        result = utils.justify("", 1, align="l")
        self.assertEqual(" ", result)

    @parameterized.expand(
        [
            (5, "Task \n ID  "),
            (6, " Task \n  ID  "),
            (7, "Task ID"),
            (8, "Task ID "),
            (9, " Task ID "),
            (10, " Task ID  "),
            (11, "  Task ID  "),
        ]
    )
    def test_center_justify_small(self, width, expected):
        result = utils.justify("Task ID", width, align="c", indent=0, fillchar=" ")
        self.assertEqual(expected, result)

    def test_justify_ansi(self):
        """
        Justify ansistring

        """

        from evennia.utils.ansi import ANSI_RED

        line = ANSIString("This is a |rred|n word")

        result = utils.justify(line, align="c", width=30)

        self.assertIn(ANSI_RED, str(result))


class TestMatchIP(TestCase):
    """
    test utils.match_ip
    """

    def test_match_ip(self):
        self.assertFalse(utils.match_ip("192.168.0.1", "10.0.0.0/24"))
        self.assertTrue(utils.match_ip("192.168.0.1", "192.168.0.0/24"))
        self.assertTrue(utils.match_ip("192.168.0.1", "192.168.0.1"))
        self.assertFalse(utils.match_ip("192.168.0.1", "10.0.0.1"))



"""

A Command Set (CmdSet) holds a set of commands. The Cmdsets can be
merged and combined to create new sets of commands in a
non-destructive way. This makes them very powerful for implementing
custom game states where different commands (or different variations
of commands) are available to the accounts depending on circumstance.

The available merge operations are partly borrowed from mathematical
Set theory.


* Union The two command sets are merged so that as many commands as
    possible of each cmdset ends up in the merged cmdset. Same-name
    commands are merged by priority.  This is the most common default.
    Ex: A1,A3 + B1,B2,B4,B5 = A1,B2,A3,B4,B5
* Intersect - Only commands found in *both* cmdsets (i.e. which have
    same names) end up in the merged cmdset, with the higher-priority
    cmdset replacing the lower one. Ex: A1,A3 + B1,B2,B4,B5 = A1
* Replace -   The commands of this cmdset completely replaces the
    lower-priority cmdset's commands, regardless of if same-name commands
    exist. Ex: A1,A3 + B1,B2,B4,B5 = A1,A3
* Remove -    This removes the relevant commands from the
    lower-priority cmdset completely.  They are not replaced with
    anything, so this in effects uses the high-priority cmdset as a filter
    to affect the low-priority cmdset.  Ex: A1,A3 + B1,B2,B4,B5 = B2,B4,B5

"""
from weakref import WeakKeyDictionary

from django.utils.translation import gettext as _

from evennia.utils.utils import inherits_from, is_iter

__all__ = ("CmdSet",)


class _CmdSetMeta(type):
    """
    This metaclass makes some minor on-the-fly convenience fixes to
    the cmdset class.

    """

    def __init__(cls, *args, **kwargs):
        """
        Fixes some things in the cmdclass

        """
        # by default we key the cmdset the same as the
        # name of its class.
        if not hasattr(cls, "key") or not cls.key:
            cls.key = cls.__name__
        cls.path = "%s.%s" % (cls.__module__, cls.__name__)

        if not isinstance(cls.key_mergetypes, dict):
            cls.key_mergetypes = {}

        super().__init__(*args, **kwargs)


class CmdSet(object, metaclass=_CmdSetMeta):
    """
    This class describes a unique cmdset that understands priorities.
    CmdSets can be merged and made to perform various set operations
    on each other.  CmdSets have priorities that affect which of their
    ingoing commands gets used.

    In the examples, cmdset A always have higher priority than cmdset B.

    key - the name of the cmdset. This can be used on its own for game
    operations

    mergetype (partly from Set theory):

        Union -    The two command sets are merged so that as many
                    commands as possible of each cmdset ends up in the
                    merged cmdset. Same-name commands are merged by
                    priority.  This is the most common default.
                    Ex: A1,A3 + B1,B2,B4,B5 = A1,B2,A3,B4,B5
        Intersect - Only commands found in *both* cmdsets
                    (i.e. which have same names) end up in the merged
                    cmdset, with the higher-priority cmdset replacing the
                    lower one.  Ex: A1,A3 + B1,B2,B4,B5 = A1
        Replace -   The commands of this cmdset completely replaces
                    the lower-priority cmdset's commands, regardless
                    of if same-name commands exist.
                    Ex: A1,A3 + B1,B2,B4,B5 = A1,A3
        Remove -    This removes the relevant commands from the
                    lower-priority cmdset completely.  They are not
                    replaced with anything, so this in effects uses the
                    high-priority cmdset as a filter to affect the
                    low-priority cmdset.
                    Ex: A1,A3 + B1,B2,B4,B5 = B2,B4,B5

                 Note: Commands longer than 2 characters and starting
                       with double underscrores, like '__noinput_command'
                       are considered 'system commands' and are
                       excempt from all merge operations - they are
                       ALWAYS included across mergers and only affected
                       if same-named system commands replace them.

    priority- All cmdsets are always merged in pairs of two so that
              the higher set's mergetype is applied to the
              lower-priority cmdset. Default commands have priority 0,
              high-priority ones like Exits and Channels have 10 and 9.
              Priorities can be negative as well to give default
              commands preference.

    duplicates - determines what happens when two sets of equal
                 priority merge (only). Defaults to None and has the first of them in the
                 merger (i.e. A above) automatically taking
                 precedence. But if `duplicates` is true, the
                 result will be a merger with more than one of each
                 name match.  This will usually lead to the account
                 receiving a multiple-match error higher up the road,
                 but can be good for things like cmdsets on non-account
                 objects in a room, to allow the system to warn that
                 more than one 'ball' in the room has the same 'kick'
                 command defined on it, so it may offer a chance to
                 select which ball to kick ...  Allowing duplicates
                 only makes sense for Union and Intersect, the setting
                 is ignored for the other mergetypes.
                 Note that the `duplicates` flag is *not* propagated in
                 a cmdset merger. So `A + B = C` will result in
                 a cmdset with duplicate commands, but C.duplicates will
                 be `None`. For duplication to apply to a whole cmdset
                 stack merge, _all_ cmdsets in the stack must have
                 `.duplicates=True` set.
                Finally, if a final cmdset has `.duplicates=None` (the normal
                unless created alone with another value), the cmdhandler
                will assume True for object-based cmdsets and False for
                all other. This is usually the most intuitive outcome.

    key_mergetype (dict) - allows the cmdset to define a unique
             mergetype for particular cmdsets.  Format is
             {CmdSetkeystring:mergetype}. Priorities still apply.
             Example: {'Myevilcmdset','Replace'} which would make
             sure for this set to always use 'Replace' on
             Myevilcmdset no matter what overall mergetype this set
             has.

    no_objs  - don't include any commands from nearby objects
                  when searching for suitable commands
    no_exits  - ignore the names of exits when matching against
                        commands
    no_channels   - ignore the name of channels when matching against
                        commands (WARNING- this is dangerous since the
                        account can then not even ask staff for help if
                        something goes wrong)


    """

    key = "Unnamed CmdSet"
    mergetype = "Union"
    priority = 0

    # These flags, if set to None should be interpreted as 'I don't care' and,
    # will allow "pass-through" even of lower-prio cmdsets' explicitly True/False
    # options. If this is set to True/False however, priority matters.
    no_exits = None
    no_objs = None
    no_channels = None
    # The .duplicates setting does not propagate and since duplicates can only happen
    # on same-prio cmdsets, there is no concept of passthrough on `None`.
    # The merger of two cmdsets always return in a cmdset with `duplicates=None`
    # (even if the result may have duplicated commands).
    # If a final cmdset has `duplicates=None` (normal, unless the cmdset is
    # created on its own with the flag set), the cmdhandler will auto-assume it to be
    # True for Object-based cmdsets and stay None/False for all other entities.
    #
    # Example:
    #  A and C has .duplicates=True, B has .duplicates=None (or False)
    #  B + A = BA, where BA will have duplicate cmds, but BA.duplicates = None
    #  BA + C = BAC, where BAC will have more duplication, but BAC.duplicates = None
    #
    # Basically, for the `.duplicate` setting to survive throughout a
    # merge-stack, every cmdset in the stack must have `duplicates` set explicitly.
    duplicates = None

    persistent = False
    key_mergetypes = {}
    errmessage = ""
    # pre-store properties to duplicate straight off
    to_duplicate = (
        "key",
        "cmdsetobj",
        "no_exits",
        "no_objs",
        "no_channels",
        "persistent",
        "mergetype",
        "priority",
        "duplicates",
        "errmessage",
    )

    def __init__(self, cmdsetobj=None, key=None):
        """
        Creates a new CmdSet instance.

        Args:
            cmdsetobj (Session, Account, Object, optional): This is the database object
                to which this particular instance of cmdset is related. It
                is often a character but may also be a regular object, Account
                or Session.
            key (str, optional): The idenfier for this cmdset. This
                helps if wanting to selectively remov cmdsets.

        """

        if key:
            self.key = key
        self.commands = []
        self.system_commands = []
        self.actual_mergetype = self.mergetype
        self.cmdsetobj = cmdsetobj
        # this is set only on merged sets, in cmdhandler.py, in order to
        # track, list and debug mergers correctly.
        self.merged_from = []

        # initialize system
        self.at_cmdset_creation()
        self._contains_cache = WeakKeyDictionary()  # {}

    # Priority-sensitive merge operations for cmdsets

    def _union(self, cmdset_a, cmdset_b):
        """
        Merge two sets using union merger

        Args:
            cmdset_a (Cmdset): Cmdset given higher priority in the case of a tie.
            cmdset_b (Cmdset): Cmdset given lower priority in the case of a tie.

        Returns:
            cmdset_c (Cmdset): The result of A U B operation.

        Notes:
            Union, C = A U B,  means that C gets all elements from both A and B.

        """
        cmdset_c = cmdset_a._duplicate()
        # we make copies, not refs by use of [:]
        cmdset_c.commands = cmdset_a.commands[:]
        if cmdset_a.duplicates and cmdset_a.priority == cmdset_b.priority:
            cmdset_c.commands.extend(cmdset_b.commands)
        else:
            cmdset_c.commands.extend([cmd for cmd in cmdset_b if cmd not in cmdset_a])
        return cmdset_c

    def _intersect(self, cmdset_a, cmdset_b):
        """
        Merge two sets using intersection merger

        Args:
            cmdset_a (Cmdset): Cmdset given higher priority in the case of a tie.
            cmdset_b (Cmdset): Cmdset given lower priority in the case of a tie.

        Returns:
            cmdset_c (Cmdset): The result of A (intersect) B operation.

        Notes:
            Intersection, C = A (intersect) B, means that C only gets the
                parts of A and B that are the same (that is, the commands
                of each set having the same name. Only the one of these
                having the higher prio ends up in C).

        """
        cmdset_c = cmdset_a._duplicate()
        if cmdset_a.duplicates and cmdset_a.priority == cmdset_b.priority:
            for cmd in [cmd for cmd in cmdset_a if cmd in cmdset_b]:
                cmdset_c.add(cmd)
                cmdset_c.add(cmdset_b.get(cmd))
        else:
            cmdset_c.commands = [cmd for cmd in cmdset_a if cmd in cmdset_b]
        return cmdset_c

    def _replace(self, cmdset_a, cmdset_b):
        """
        Replace the contents of one set with another

        Args:
            cmdset_a (Cmdset): Cmdset replacing
            cmdset_b (Cmdset): Cmdset to replace

        Returns:
            cmdset_c (Cmdset): This is indentical to cmdset_a.

        Notes:
            C = A, where B is ignored.

        """
        cmdset_c = cmdset_a._duplicate()
        cmdset_c.commands = cmdset_a.commands[:]
        return cmdset_c

    def _remove(self, cmdset_a, cmdset_b):
        """
        Filter a set by another.

        Args:
            cmdset_a (Cmdset): Cmdset acting as a removal filter.
            cmdset_b (Cmdset): Cmdset to filter

        Returns:
            cmdset_c (Cmdset): B, with all matching commands from A removed.

        Notes:
            C = B - A, where A is used to remove the commands of B.

        """

        cmdset_c = cmdset_a._duplicate()
        cmdset_c.commands = [cmd for cmd in cmdset_b if cmd not in cmdset_a]
        return cmdset_c

    def _instantiate(self, cmd):
        """
        checks so that object is an instantiated command and not, say
        a cmdclass. If it is, instantiate it.  Other types, like
        strings, are passed through.

        Args:
            cmd (any): Entity to analyze.

        Returns:
            result (any): An instantiated Command or the input unmodified.

        """
        if callable(cmd):
            return cmd()
        else:
            return cmd

    def _duplicate(self):
        """
        Returns a new cmdset with the same settings as this one (no
        actual commands are copied over)

        Returns:
            cmdset (Cmdset): A copy of the current cmdset.
        """
        cmdset = CmdSet()
        for key, val in ((key, getattr(self, key)) for key in self.to_duplicate):
            if val != getattr(cmdset, key):
                # only copy if different from default; avoid turning
                # class-vars into instance vars
                setattr(cmdset, key, val)
        cmdset.key_mergetypes = self.key_mergetypes.copy()
        return cmdset

    def __str__(self):
        """
        Show all commands in cmdset when printing it.

        Returns:
            commands (str): Representation of commands in Cmdset.

        """
        perm = "perm" if self.persistent else "non-perm"
        options = ", ".join(
            [
                "{}:{}".format(opt, "T" if getattr(self, opt) else "F")
                for opt in ("no_exits", "no_objs", "no_channels", "duplicates")
                if getattr(self, opt) is not None
            ]
        )
        options = (", " + options) if options else ""
        return (
            f"<CmdSet {self.key}, {self.mergetype}, {perm}, prio {self.priority}{options}>: "
            + ", ".join([str(cmd) for cmd in sorted(self.commands, key=lambda o: o.key)])
        )

    def __iter__(self):
        """
        Allows for things like 'for cmd in cmdset':

        Returns:
            iterable (iter): Commands in Cmdset.

        """
        return iter(self.commands)

    def __contains__(self, othercmd):
        """
        Returns True if this cmdset contains the given command (as
        defined by command name and aliases). This allows for things
        like 'if cmd in cmdset'

        """
        ret = self._contains_cache.get(othercmd)
        if ret is None:
            ret = othercmd in self.commands
            self._contains_cache[othercmd] = ret
        return ret

    def __add__(self, cmdset_a):
        """
        Merge this cmdset (B) with another cmdset (A) using the + operator,

        C = B + A

        Here, we (by convention) say that 'A is merged onto B to form
        C'.  The actual merge operation used in the 'addition' depends
        on which priorities A and B have. The one of the two with the
        highest priority will apply and give its properties to C. In
        the case of a tie, A  takes priority and replaces the
        same-named commands in B unless A has the 'duplicate' variable
        set (which means both sets' commands are kept).
        """

        # It's okay to merge with None
        if not cmdset_a:
            return self

        sys_commands_a = cmdset_a.get_system_cmds()
        sys_commands_b = self.get_system_cmds()

        if self.priority <= cmdset_a.priority:
            # A higher or equal priority to B

            # preserve system __commands
            sys_commands = sys_commands_a + [
                cmd for cmd in sys_commands_b if cmd not in sys_commands_a
            ]

            mergetype = cmdset_a.key_mergetypes.get(self.key, cmdset_a.mergetype)
            if mergetype == "Intersect":
                cmdset_c = self._intersect(cmdset_a, self)
            elif mergetype == "Replace":
                cmdset_c = self._replace(cmdset_a, self)
            elif mergetype == "Remove":
                cmdset_c = self._remove(cmdset_a, self)
            else:  # Union
                cmdset_c = self._union(cmdset_a, self)

            # pass through options whenever they are set, unless the merging or higher-prio
            # set changes the setting (i.e. has a non-None value). We don't pass through
            # the duplicates setting; that is per-merge; the resulting .duplicates value
            # is always None (so merging cmdsets must all have explicit values if wanting
            # to cause duplicates).
            cmdset_c.no_channels = (
                self.no_channels if cmdset_a.no_channels is None else cmdset_a.no_channels
            )
            cmdset_c.no_exits = self.no_exits if cmdset_a.no_exits is None else cmdset_a.no_exits
            cmdset_c.no_objs = self.no_objs if cmdset_a.no_objs is None else cmdset_a.no_objs
            cmdset_c.duplicates = None

        else:
            # B higher priority than A

            # preserver system __commands
            sys_commands = sys_commands_b + [
                cmd for cmd in sys_commands_a if cmd not in sys_commands_b
            ]

            mergetype = self.key_mergetypes.get(cmdset_a.key, self.mergetype)
            if mergetype == "Intersect":
                cmdset_c = self._intersect(self, cmdset_a)
            elif mergetype == "Replace":
                cmdset_c = self._replace(self, cmdset_a)
            elif mergetype == "Remove":
                cmdset_c = self._remove(self, cmdset_a)
            else:  # Union
                cmdset_c = self._union(self, cmdset_a)

            # pass through options whenever they are set, unless the higher-prio
            # set changes the setting (i.e. has a non-None value). We don't pass through
            # the duplicates setting; that is per-merge; the resulting .duplicates value#
            # is always None (so merging cmdsets must all have explicit values if wanting
            # to cause duplicates).
            cmdset_c.no_channels = (
                cmdset_a.no_channels if self.no_channels is None else self.no_channels
            )
            cmdset_c.no_exits = cmdset_a.no_exits if self.no_exits is None else self.no_exits
            cmdset_c.no_objs = cmdset_a.no_objs if self.no_objs is None else self.no_objs
            cmdset_c.duplicates = None

        # we store actual_mergetype since key_mergetypes
        # might be different from the main mergetype.
        # This is used for diagnosis.
        cmdset_c.actual_mergetype = mergetype

        # print "__add__ for %s (prio %i)  called with %s (prio %i)." % (self.key, self.priority,
        # cmdset_a.key, cmdset_a.priority)

        # return the system commands to the cmdset
        cmdset_c.add(sys_commands, allow_duplicates=True)
        return cmdset_c

    def add(self, cmd, allow_duplicates=False):
        """
        Add a new command or commands to this CmdSet, a list of
        commands or a cmdset to this cmdset. Note that this is *not*
        a merge operation (that is handled by the + operator).

        Args:
            cmd (Command, list, Cmdset): This allows for adding one or
                more commands to this Cmdset in one go. If another Cmdset
                is given, all its commands will be added.
            allow_duplicates (bool, optional): If set, will not try to remove
                duplicate cmds in the set. This is needed during the merge process
                to avoid wiping commands coming from cmdsets with duplicate=True.

        Notes:
            If cmd already exists in set, it will replace the old one
            (no priority checking etc happens here). This is very useful
            when overloading default commands).

            If cmd is another cmdset class or -instance, the commands of
            that command set is added to this one, as if they were part of
            the original cmdset definition. No merging or priority checks
            are made, rather later added commands will simply replace
            existing ones to make a unique set.

        """
        if inherits_from(cmd, "evennia.commands.cmdset.CmdSet"):
            # cmd is a command set so merge all commands in that set
            # to this one. We raise a visible error if we created
            # an infinite loop (adding cmdset to itself somehow)
            cmdset = cmd
            try:
                cmdset = self._instantiate(cmdset)
            except RuntimeError:
                err = (
                    "Adding cmdset {cmdset} to {cls} lead to an "
                    "infinite loop. When adding a cmdset to another, "
                    "make sure they are not themself cyclically added to "
                    "the new cmdset somewhere in the chain."
                )
                raise RuntimeError(_(err.format(cmdset=cmdset, cls=self.__class__)))
            cmds = cmdset.commands
        elif is_iter(cmd):
            cmds = [self._instantiate(c) for c in cmd]
        else:
            cmds = [self._instantiate(cmd)]

        commands = self.commands
        system_commands = self.system_commands

        for cmd in cmds:
            # Ensure commands know their source cmdset.
            cmd.cmdset_source = self
            # add all commands
            if not hasattr(cmd, "obj") or cmd.obj is None:
                cmd.obj = self.cmdsetobj

            # remove duplicates and add new
            for _dum in range(commands.count(cmd)):
                commands.remove(cmd)
            commands.append(cmd)

            # add system_command to separate list as well,
            # for quick look-up. These have no
            if cmd.key.startswith("__"):
                # remove same-matches and add new
                for _dum in range(system_commands.count(cmd)):
                    system_commands.remove(cmd)
                system_commands.append(cmd)

        if not allow_duplicates:
            # extra run to make sure to avoid doublets
            commands = list(set(commands))
        self.commands = commands

    def remove(self, cmd):
        """
        Remove a command instance from the cmdset.

        Args:
            cmd (Command or str): Either the Command object to remove
                or the key of such a command.

        """
        if isinstance(cmd, str):
            _cmd = next((_cmd for _cmd in self.commands if _cmd.key == cmd), None)
            if _cmd is None:
                if not cmd.startswith("__"):
                    # if a syscommand, keep the original string and instantiate on it
                    return None
            else:
                cmd = _cmd

        cmd = self._instantiate(cmd)
        if cmd.key.startswith("__"):
            try:
                ic = self.system_commands.index(cmd)
                del self.system_commands[ic]
            except ValueError:
                # ignore error
                pass
        else:
            self.commands = [oldcmd for oldcmd in self.commands if oldcmd != cmd]

    def get(self, cmd):
        """
        Get a command from the cmdset. This is mostly useful to
        check if the command is part of this cmdset or not.

        Args:
            cmd (Command or str): Either the Command object or its key.

        Returns:
            cmd (Command): The first matching Command in the set.

        """
        if isinstance(cmd, str):
            _cmd = next((_cmd for _cmd in self.commands if _cmd.key == cmd), None)
            if _cmd is None:
                if not cmd.startswith("__"):
                    # if a syscommand, keep the original string and instantiate on it
                    return None
            else:
                cmd = _cmd

        cmd = self._instantiate(cmd)
        for thiscmd in self.commands:
            if thiscmd == cmd:
                return thiscmd
        return None

    def count(self):
        """
        Number of commands in set.

        Returns:
            N (int): Number of commands in this Cmdset.

        """
        return len(self.commands)

    def get_system_cmds(self):
        """
        Get system commands in cmdset

        Returns:
            sys_cmds (list): The system commands in the set.

        Notes:
            As far as the Cmdset is concerned, system commands are any
            commands with a key starting with double underscore __.
            These are excempt from merge operations.

        """
        return self.system_commands

    def make_unique(self, caller):
        """
        Remove duplicate command-keys (unsafe)

        Args:
            caller (object): Commands on this object will
                get preference in the duplicate removal.

        Notes:
            This is an unsafe command meant to clean out a cmdset of
            doublet commands after it has been created. It is useful
            for commands inheriting cmdsets from the cmdhandler where
            obj-based cmdsets always are added double. Doublets will
            be weeded out with preference to commands defined on
            caller, otherwise just by first-come-first-served.

        """
        unique = {}
        for cmd in self.commands:
            if cmd.key in unique:
                ocmd = unique[cmd.key]
                if (hasattr(cmd, "obj") and cmd.obj == caller) and not (
                    hasattr(ocmd, "obj") and ocmd.obj == caller
                ):
                    unique[cmd.key] = cmd
            else:
                unique[cmd.key] = cmd
        self.commands = list(unique.values())

    def get_all_cmd_keys_and_aliases(self, caller=None):
        """
        Collects keys/aliases from commands

        Args:
            caller (Object, optional): If set, this is used to check access permissions
                on each command. Only commands that pass are returned.

        Returns:
            names (list): A list of all command keys and aliases in this cmdset. If `caller`
                was given, this list will only contain commands to which `caller` passed
                the `call` locktype check.

        """
        names = []
        if caller:
            [names.extend(cmd._keyaliases) for cmd in self.commands if cmd.access(caller)]
        else:
            [names.extend(cmd._keyaliases) for cmd in self.commands]
        return names

    def at_cmdset_creation(self):
        """
        Hook method - this should be overloaded in the inheriting
        class, and should take care of populating the cmdset by use of
        self.add().

        """
        pass



"""
Module containing the task handler for Evennia deferred tasks, persistent or not.
"""

from datetime import datetime, timedelta
from pickle import PickleError

from twisted.internet import reactor
from twisted.internet.defer import CancelledError as DefCancelledError
from twisted.internet.task import deferLater

from evennia.server.models import ServerConfig
from evennia.utils.dbserialize import dbserialize, dbunserialize
from evennia.utils.logger import log_err

TASK_HANDLER = None


def handle_error(*args, **kwargs):
    """Handle errors within deferred objects."""
    for arg in args:
        # suppress cancel errors
        if arg.type == DefCancelledError:
            continue
        raise arg


class TaskHandlerTask:
    """An object to represent a single TaskHandler task.

    Instance Attributes:
        task_id (int): the global id for this task
        deferred (deferred): a reference to this task's deferred
    Property Attributes:
        paused (bool): check if the deferred instance of a task has been paused.
        called(self): A task attribute to check if the deferred instance of a task has been called.

    Methods:
        pause(): Pause the callback of a task.
        unpause(): Process all callbacks made since pause() was called.
        do_task(): Execute the task (call its callback).
        call(): Call the callback of this task.
        remove(): Remove a task without executing it.
        cancel(): Stop a task from automatically executing.
        active(): Check if a task is active (has not been called yet).
        exists(): Check if a task exists.
        get_id(): Returns the global id for this task. For use with

    """

    def __init__(self, task_id):
        self.task_id = task_id
        self.deferred = TASK_HANDLER.get_deferred(task_id)

    def get_deferred(self):
        """Return the instance of the deferred the task id is using.

        Returns:
            bool or deferred: An instance of a deferred or False if there is no task with the id.
                None is returned if there is no deferred affiliated with this id.

        """
        return TASK_HANDLER.get_deferred(self.task_id)

    def pause(self):
        """
        Pause the callback of a task.
        To resume use `TaskHandlerTask.unpause`.

        """
        d = self.deferred
        if d:
            d.pause()

    def unpause(self):
        """
        Unpause a task, run the task if it has passed delay time.

        """
        d = self.deferred
        if d:
            d.unpause()

    @property
    def paused(self):
        """
        A task attribute to check if the deferred instance of a task has been paused.

        This exists to mock usage of a twisted deferred object.

        Returns:
            bool or None: True if the task was properly paused. None if the task does not have
                a deferred instance.

        """
        d = self.deferred
        if d:
            return d.paused
        else:
            return None

    def do_task(self):
        """
        Execute the task (call its callback).
        If calling before timedelay, cancel the deferred instance affliated to this task.
        Remove the task from the dictionary of current tasks on a successful
        callback.

        Returns:
            bool or any: Set to `False` if the task does not exist in task
            handler. Otherwise it will be the return of the task's callback.

        """
        return TASK_HANDLER.do_task(self.task_id)

    def call(self):
        """
        Call the callback of a task.
        Leave the task unaffected otherwise.
        This does not use the task's deferred instance.
        The only requirement is that the task exist in task handler.

        Returns:
            bool or any: Set to `False` if the task does not exist in task
            handler. Otherwise it will be the return of the task's callback.

        """
        return TASK_HANDLER.call_task(self.task_id)

    def remove(self):
        """Remove a task without executing it.
        Deletes the instance of the task's deferred.

        Args:
            task_id (int): an existing task ID.

        Returns:
            bool: True if the removal completed successfully.

        """
        return TASK_HANDLER.remove(self.task_id)

    def cancel(self):
        """Stop a task from automatically executing.
        This will not remove the task.

        Returns:
            bool: True if the cancel completed successfully.
                False if the cancel did not complete successfully.

        """
        return TASK_HANDLER.cancel(self.task_id)

    def active(self):
        """Check if a task is active (has not been called yet).

        Returns:
            bool: True if a task is active (has not been called yet). False if
                it is not (has been called) or if the task does not exist.

        """
        return TASK_HANDLER.active(self.task_id)

    @property
    def called(self):
        """
        A task attribute to check if the deferred instance of a task has been called.

        This exists to mock usage of a twisted deferred object.
        It will not set to True if Task.call has been called. This only happens if
        task's deferred instance calls the callback.

        Returns:
            bool: True if the deferred instance of this task has called the callback.
                False if the deferred instnace of this task has not called the callback.

        """
        d = self.deferred
        if d:
            return d.called
        else:
            return None

    def exists(self):
        """
        Check if a task exists.
        Most task handler methods check for existence for you.

        Returns:
            bool: True the task exists False if it does not.

        """
        return TASK_HANDLER.exists(self.task_id)

    def get_id(self):
        """
        Returns the global id for this task. For use with
        `evennia.scripts.taskhandler.TASK_HANDLER`.

        Returns:
            task_id (int): global task id for this task.

        """
        return self.task_id


class TaskHandler:

    """A light singleton wrapper allowing to access permanent tasks.

    When `utils.delay` is called, the task handler is used to create
    the task.

    Task handler will automatically remove uncalled but canceled from task
    handler. By default this will not occur until a canceled task
    has been uncalled for 60 second after the time it should have been called.
    To adjust this time use TASK_HANDLER.stale_timeout. If stale_timeout is 0
    stale tasks will not be automatically removed.
    This is not done on a timer. I is done as new tasks are added or the load method is called.

    """

    def __init__(self):
        self.tasks = {}
        self.to_save = {}
        self.clock = reactor
        # number of seconds before an uncalled canceled task is removed from TaskHandler
        self.stale_timeout = 60
        self._now = False  # used in unit testing to manually set now time

    def load(self):
        """Load from the ServerConfig.

        This should be automatically called when Evennia starts.
        It populates `self.tasks` according to the ServerConfig.

        """
        to_save = False
        value = ServerConfig.objects.conf("delayed_tasks", default=dict)
        if isinstance(value, str):
            tasks = dbunserialize(value)
        else:
            tasks = value

        # At this point, `tasks` contains a dictionary of still-serialized tasks
        for task_id, value in tasks.items():
            date, callback, args, kwargs = dbunserialize(value)
            if isinstance(callback, tuple):
                # `callback` can be an object and name for instance methods
                obj, method = callback
                if obj is None:
                    to_save = True
                    continue

                callback = getattr(obj, method)
            self.tasks[task_id] = (date, callback, args, kwargs, True, None)

        if self.stale_timeout > 0:  # cleanup stale tasks.
            self.clean_stale_tasks()
        if to_save:
            self.save()

    def clean_stale_tasks(self):
        """remove uncalled but canceled from task handler.

        By default this will not occur until a canceled task
        has been uncalled for 60 second after the time it should have been called.
        To adjust this time use TASK_HANDLER.stale_timeout.

        """
        clean_ids = []
        for task_id, (date, callback, args, kwargs, persistent, _) in self.tasks.items():
            if not self.active(task_id):
                stale_date = date + timedelta(seconds=self.stale_timeout)
                # if a now time is provided use it (intended for unit testing)
                now = self._now if self._now else datetime.now()
                # the task was canceled more than stale_timeout seconds ago
                if now > stale_date:
                    clean_ids.append(task_id)
        for task_id in clean_ids:
            self.remove(task_id)
        return True

    def save(self):
        """
        Save the tasks in ServerConfig.

        """

        for task_id, (date, callback, args, kwargs, persistent, _) in self.tasks.items():
            if task_id in self.to_save:
                continue
            if not persistent:
                continue

            safe_callback = callback
            if getattr(callback, "__self__", None):
                # `callback` is an instance method
                obj = callback.__self__
                name = callback.__name__
                safe_callback = (obj, name)

            # Check if callback can be pickled. args and kwargs have been checked
            try:
                dbserialize(safe_callback)
            except (TypeError, AttributeError, PickleError) as err:
                raise ValueError(
                    "the specified callback {callback} cannot be pickled. "
                    "It must be a top-level function in a module or an "
                    "instance method ({err}).".format(callback=callback, err=err)
                )

            self.to_save[task_id] = dbserialize((date, safe_callback, args, kwargs))

        ServerConfig.objects.conf("delayed_tasks", self.to_save)

    def add(self, timedelay, callback, *args, **kwargs):
        """
        Add a new task.

        If the persistent kwarg is truthy:
        The callback, args and values for kwarg will be serialized. Type
        and attribute errors during the serialization will be logged,
        but will not throw exceptions.
        For persistent tasks do not use memory references in the callback
        function or arguments. After a restart those memory references are no
        longer accurate.

        Args:
            timedelay (int or float): time in seconds before calling the callback.
            callback (function or instance method): the callback itself
            any (any): any additional positional arguments to send to the callback
            *args: positional arguments to pass to callback.
            **kwargs: keyword arguments to pass to callback.
                - persistent (bool, optional): persist the task (stores it).
                    Persistent key and value is removed from kwargs it will
                    not be passed to callback.

        Returns:
            TaskHandlerTask: An object to represent a task.
                Reference `evennia.scripts.taskhandler.TaskHandlerTask` for complete details.

        """
        # set the completion time
        # Only used on persistent tasks after a restart
        now = datetime.now()
        delta = timedelta(seconds=timedelay)
        comp_time = now + delta
        # get an open task id
        used_ids = list(self.tasks.keys())
        task_id = 1
        while task_id in used_ids:
            task_id += 1

        # record the task to the tasks dictionary
        persistent = kwargs.get("persistent", False)
        if "persistent" in kwargs:
            del kwargs["persistent"]
        if persistent:
            safe_args = []
            safe_kwargs = {}

            # Check that args and kwargs contain picklable information
            for arg in args:
                try:
                    dbserialize(arg)
                except (TypeError, AttributeError, PickleError):
                    log_err(
                        "The positional argument {} cannot be "
                        "pickled and will not be present in the arguments "
                        "fed to the callback {}".format(arg, callback)
                    )
                else:
                    safe_args.append(arg)

            for key, value in kwargs.items():
                try:
                    dbserialize(value)
                except (TypeError, AttributeError, PickleError):
                    log_err(
                        "The {} keyword argument {} cannot be "
                        "pickled and will not be present in the arguments "
                        "fed to the callback {}".format(key, value, callback)
                    )
                else:
                    safe_kwargs[key] = value

            self.tasks[task_id] = (comp_time, callback, safe_args, safe_kwargs, persistent, None)
            self.save()
        else:  # this is a non-persitent task
            self.tasks[task_id] = (comp_time, callback, args, kwargs, persistent, None)

        # defer the task
        callback = self.do_task
        args = [task_id]
        kwargs = {}
        d = deferLater(self.clock, timedelay, callback, *args, **kwargs)
        d.addErrback(handle_error)

        # some tasks may complete before the deferred can be added
        if task_id in self.tasks:
            task = self.tasks.get(task_id)
            task = list(task)
            task[4] = persistent
            task[5] = d
            self.tasks[task_id] = task
        else:  # the task already completed
            return False
        if self.stale_timeout > 0:
            self.clean_stale_tasks()
        return TaskHandlerTask(task_id)

    def exists(self, task_id):
        """
        Check if a task exists.
        Most task handler methods check for existence for you.

        Args:
            task_id (int): an existing task ID.

        Returns:
            bool: True the task exists False if it does not.

        """
        if task_id in self.tasks:
            return True
        else:
            return False

    def active(self, task_id):
        """
        Check if a task is active (has not been called yet).

        Args:
            task_id (int): an existing task ID.

        Returns:
            bool: True if a task is active (has not been called yet). False if
                it is not (has been called) or if the task does not exist.

        """
        if task_id in self.tasks:
            # if the task has not been run, cancel it
            deferred = self.get_deferred(task_id)
            return not (deferred and deferred.called)
        else:
            return False

    def cancel(self, task_id):
        """
        Stop a task from automatically executing.
        This will not remove the task.

        Args:
            task_id (int): an existing task ID.

        Returns:
            bool: True if the cancel completed successfully.
                False if the cancel did not complete successfully.

        """
        if task_id in self.tasks:
            # if the task has not been run, cancel it
            d = self.get_deferred(task_id)
            if d:  # it is remotely possible for a task to not have a deferred
                if d.called:
                    return False
                else:  # the callback has not been called yet.
                    d.cancel()
                    return True
            else:  # this task has no deferred instance
                return False
        else:
            return False

    def remove(self, task_id):
        """
        Remove a task without executing it.
        Deletes the instance of the task's deferred.

        Args:
            task_id (int): an existing task ID.

        Returns:
            bool: True if the removal completed successfully.

        """
        d = None
        # delete the task from the tasks dictionary
        if task_id in self.tasks:
            # if the task has not been run, cancel it
            self.cancel(task_id)
            del self.tasks[task_id]  # delete the task from the tasks dictionary
        # remove the task from the persistent dictionary and ServerConfig
        if task_id in self.to_save:
            del self.to_save[task_id]
            self.save()  # remove from ServerConfig.objects
        # delete the instance of the deferred
        if d:
            del d
        return True

    def clear(self, save=True, cancel=True):
        """
        Clear all tasks. By default tasks are canceled and removed from the database as well.

        Args:
            save=True (bool): Should changes to persistent tasks be saved to database.
            cancel=True (bool): Cancel scheduled tasks before removing it from task handler.

        Returns:
            True (bool): if the removal completed successfully.

        """
        if self.tasks:
            for task_id in self.tasks.keys():
                if cancel:
                    self.cancel(task_id)
            self.tasks = {}
        if self.to_save:
            self.to_save = {}
        if save:
            self.save()
        return True

    def call_task(self, task_id):
        """
        Call the callback of a task.
        Leave the task unaffected otherwise.
        This does not use the task's deferred instance.
        The only requirement is that the task exist in task handler.

        Args:
            task_id (int): an existing task ID.

        Returns:
            bool or any: Set to `False` if the task does not exist in task
            handler. Otherwise it will be the return of the task's callback.

        """
        if task_id in self.tasks:
            date, callback, args, kwargs, persistent, d = self.tasks.get(task_id)
        else:  # the task does not exist
            return False
        return callback(*args, **kwargs)

    def do_task(self, task_id):
        """
        Execute the task (call its callback).
        If calling before timedelay cancel the deferred instance affliated to this task.
        Remove the task from the dictionary of current tasks on a successful
        callback.

        Args:
            task_id (int): a valid task ID.

        Returns:
            bool or any: Set to `False` if the task does not exist in task
            handler. Otherwise it will be the return of the task's callback.

        """
        callback_return = False
        if task_id in self.tasks:
            date, callback, args, kwargs, persistent, d = self.tasks.get(task_id)
        else:  # the task does not exist
            return False
        if d:  # it is remotely possible for a task to not have a deferred
            if not d.called:  # the task's deferred has not been called yet
                d.cancel()  # cancel the automated callback
        else:  # this task has no deferred, and should not be called
            return False
        callback_return = callback(*args, **kwargs)
        self.remove(task_id)
        return callback_return

    def get_deferred(self, task_id):
        """
        Return the instance of the deferred the task id is using.

        Args:
            task_id (int): a valid task ID.

        Returns:
            bool or deferred: An instance of a deferred or False if there is no task with the id.
                None is returned if there is no deferred affiliated with this id.

        """
        if task_id in self.tasks:
            return self.tasks[task_id][5]
        else:
            return None

    def create_delays(self):
        """
        Create the delayed tasks for the persistent tasks.
        This method should be automatically called when Evennia starts.

        """
        now = datetime.now()
        for task_id, (date, callback, args, kwargs, _, _) in self.tasks.items():
            self.tasks[task_id] = date, callback, args, kwargs, True, None
            seconds = max(0, (date - now).total_seconds())
            d = deferLater(self.clock, seconds, self.do_task, task_id)
            d.addErrback(handle_error)
            # some tasks may complete before the deferred can be added
            if self.tasks.get(task_id, False):
                self.tasks[task_id] = date, callback, args, kwargs, True, d


# Create the soft singleton
TASK_HANDLER = TaskHandler()



# -*- coding: utf-8 -*-


import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("typeclasses", "0002_auto_20150109_0913")]

    operations = [
        migrations.CreateModel(
            name="ObjectDB",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID", serialize=False, auto_created=True, primary_key=True
                    ),
                ),
                ("db_key", models.CharField(max_length=255, verbose_name="key", db_index=True)),
                (
                    "db_typeclass_path",
                    models.CharField(
                        help_text="this defines what 'type' of entity this is. This variable holds a Python path to a module with a valid Evennia Typeclass.",
                        max_length=255,
                        null=True,
                        verbose_name="typeclass",
                    ),
                ),
                (
                    "db_date_created",
                    models.DateTimeField(auto_now_add=True, verbose_name="creation date"),
                ),
                (
                    "db_lock_storage",
                    models.TextField(
                        help_text="locks limit access to an entity. A lock is defined as a 'lock string' on the form 'type:lockfunctions', defining what functionality is locked and how to determine access. Not defining a lock means no access is granted.",
                        verbose_name="locks",
                        blank=True,
                    ),
                ),
                (
                    "db_sessid",
                    models.CommaSeparatedIntegerField(
                        help_text="csv list of session ids of connected Account, if any.",
                        max_length=32,
                        null=True,
                        verbose_name="session id",
                    ),
                ),
                (
                    "db_cmdset_storage",
                    models.CharField(
                        help_text="optional python path to a cmdset class.",
                        max_length=255,
                        null=True,
                        verbose_name="cmdset",
                        blank=True,
                    ),
                ),
                (
                    "db_attributes",
                    models.ManyToManyField(
                        help_text="attributes on this object. An attribute can hold any pickle-able python object (see docs for special cases).",
                        to="typeclasses.Attribute",
                        null=True,
                    ),
                ),
                (
                    "db_destination",
                    models.ForeignKey(
                        related_name="destinations_set",
                        on_delete=django.db.models.deletion.SET_NULL,
                        blank=True,
                        to="objects.ObjectDB",
                        help_text="a destination, used only by exit objects.",
                        null=True,
                        verbose_name="destination",
                    ),
                ),
                (
                    "db_home",
                    models.ForeignKey(
                        related_name="homes_set",
                        on_delete=django.db.models.deletion.SET_NULL,
                        verbose_name="home location",
                        blank=True,
                        to="objects.ObjectDB",
                        null=True,
                    ),
                ),
                (
                    "db_location",
                    models.ForeignKey(
                        related_name="locations_set",
                        on_delete=django.db.models.deletion.SET_NULL,
                        verbose_name="game location",
                        blank=True,
                        to="objects.ObjectDB",
                        null=True,
                    ),
                ),
            ],
            options={"verbose_name": "Object", "verbose_name_plural": "Objects"},
            bases=(models.Model,),
        )
    ]



"""
This defines how to edit help entries in Admin.
"""
from django import forms
from django.contrib import admin

from evennia.help.models import HelpEntry

from .tags import TagInline


class HelpTagInline(TagInline):
    model = HelpEntry.db_tags.through
    related_field = "helpentry"


class HelpEntryForm(forms.ModelForm):
    "Defines how to display the help entry"

    class Meta:
        model = HelpEntry
        fields = "__all__"

    db_help_category = forms.CharField(
        label="Help category", initial="General", help_text="organizes help entries in lists"
    )
    db_lock_storage = forms.CharField(
        label="Locks",
        initial="view:all()",
        required=False,
        widget=forms.Textarea(attrs={"cols": "100", "rows": "2"}),
        help_text="Set lock to view:all() unless you want it to only show to certain users."
        "<BR>Use the `edit:` limit if wanting to limit who can edit from in-game. By default it's "
        "only limited to who can use the `sethelp` command (Builders).",
    )


@admin.register(HelpEntry)
class HelpEntryAdmin(admin.ModelAdmin):
    "Sets up the admin manaager for help entries"
    inlines = [HelpTagInline]
    list_display = ("id", "db_key", "db_help_category", "db_lock_storage", "db_date_created")
    list_display_links = ("id", "db_key")
    search_fields = ["^db_key", "db_entrytext"]
    ordering = ["db_help_category", "db_key"]
    list_filter = ["db_help_category"]
    save_as = True
    save_on_top = True
    list_select_related = True
    view_on_site = False

    form = HelpEntryForm
    fieldsets = (
        (
            None,
            {
                "fields": (
                    ("db_key", "db_help_category"),
                    "db_entrytext",
                    "db_lock_storage",
                    # "db_date_created",
                ),
            },
        ),
    )



"""
Easy menu selection tree

Contrib - Tim Ashley Jenkins 2017

This module allows you to create and initialize an entire branching EvMenu
instance with nothing but a multi-line string passed to one function.

EvMenu is incredibly powerful and flexible, but using it for simple menus
can often be fairly cumbersome - a simple menu that can branch into five
categories would require six nodes, each with options represented as a list
of dictionaries.

This module provides a function, init_tree_selection, which acts as a frontend
for EvMenu, dynamically sourcing the options from a multi-line string you provide.
For example, if you define a string as such:

    TEST_MENU = '''Foo
    Bar
    Baz
    Qux'''

And then use TEST_MENU as the 'treestr' source when you call init_tree_selection
on a player:

    init_tree_selection(TEST_MENU, caller, callback)

The player will be presented with an EvMenu, like so:

    ___________________________

    Make your selection:
    ___________________________

    Foo
    Bar
    Baz
    Qux

Making a selection will pass the selection's key to the specified callback as a
string along with the caller, as well as the index of the selection (the line number
on the source string) along with the source string for the tree itself.

In addition to specifying selections on the menu, you can also specify categories.
Categories are indicated by putting options below it preceded with a '-' character.
If a selection is a category, then choosing it will bring up a new menu node, prompting
the player to select between those options, or to go back to the previous menu. In
addition, categories are marked by default with a '[+]' at the end of their key. Both
this marker and the option to go back can be disabled.

Categories can be nested in other categories as well - just go another '-' deeper. You
can do this as many times as you like. There's no hard limit to the number of
categories you can go down.

For example, let's add some more options to our menu, turning 'Bar' into a category.

    TEST_MENU = '''Foo
    Bar
    -You've got to know
    --When to hold em
    --When to fold em
    --When to walk away
    Baz
    Qux'''

Now when we call the menu, we can see that 'Bar' has become a category instead of a
selectable option.

    _______________________________

    Make your selection:
    _______________________________

    Foo
    Bar [+]
    Baz
    Qux

Note the [+] next to 'Bar'. If we select 'Bar', it'll show us the option listed under it.

    ________________________________________________________________

    Bar
    ________________________________________________________________

    You've got to know [+]
    << Go Back: Return to the previous menu.

Just the one option, which is a category itself, and the option to go back, which will
take us back to the previous menu. Let's select 'You've got to know'.

    ________________________________________________________________

    You've got to know
    ________________________________________________________________

    When to hold em
    When to fold em
    When to walk away
    << Go Back: Return to the previous menu.

Now we see the three options listed under it, too. We can select one of them or use 'Go
Back' to return to the 'Bar' menu we were just at before. It's very simple to make a
branching tree of selections!

One last thing - you can set the descriptions for the various options simply by adding a
':' character followed by the description to the option's line. For example, let's add a
description to 'Baz' in our menu:

    TEST_MENU = '''Foo
    Bar
    -You've got to know
    --When to hold em
    --When to fold em
    --When to walk away
    Baz: Look at this one: the best option.
    Qux'''

Now we see that the Baz option has a description attached that's separate from its key:

    _______________________________________________________________

    Make your selection:
    _______________________________________________________________

    Foo
    Bar [+]
    Baz: Look at this one: the best option.
    Qux

Once the player makes a selection - let's say, 'Foo' - the menu will terminate and call
your specified callback with the selection, like so:

    callback(caller, TEST_MENU, 0, "Foo")

The index of the selection is given along with a string containing the selection's key.
That way, if you have two selections in the menu with the same key, you can still
differentiate between them.

And that's all there is to it! For simple branching-tree selections, using this system is
much easier than manually creating EvMenu nodes. It also makes generating menus with dynamic
options much easier - since the source of the menu tree is just a string, you could easily
generate that string procedurally before passing it to the init_tree_selection function.
For example, if a player casts a spell or does an attack without specifying a target, instead
of giving them an error, you could present them with a list of valid targets to select by
generating a multi-line string of targets and passing it to init_tree_selection, with the
callable performing the maneuver once a selection is made.

This selection system only works for simple branching trees - doing anything really complicated
like jumping between categories or prompting for arbitrary input would still require a full
EvMenu implementation. For simple selections, however, I'm sure you will find using this function
to be much easier!

Included in this module is a sample menu and function which will let a player change the color
of their name - feel free to mess with it to get a feel for how this system works by importing
this module in your game's default_cmdsets.py module and adding CmdNameColor to your default
character's command set.

"""

from evennia import Command
from evennia.utils import evmenu
from evennia.utils.logger import log_trace


def init_tree_selection(
    treestr,
    caller,
    callback,
    index=None,
    mark_category=True,
    go_back=True,
    cmd_on_exit="look",
    start_text="Make your selection:",
):
    """
    Prompts a player to select an option from a menu tree given as a multi-line string.

    Args:
        treestr (str): Multi-lne string representing menu options
        caller (obj): Player to initialize the menu for
        callback (callable): Function to run when a selection is made. Must take 4 args:
            caller (obj): Caller given above
            treestr (str): Menu tree string given above
            index (int): Index of final selection
            selection (str): Key of final selection

    Options:
        index (int or None): Index to start the menu at, or None for top level
        mark_category (bool): If True, marks categories with a [+] symbol in the menu
        go_back (bool): If True, present an option to go back to previous categories
        start_text (str): Text to display at the top level of the menu
        cmd_on_exit(str): Command to enter when the menu exits - 'look' by default


    Notes:
        This function will initialize an instance of EvMenu with options generated
        dynamically from the source string, and passes the menu user's selection to
        a function of your choosing. The EvMenu is made of a single, repeating node,
        which will call itself over and over at different levels of the menu tree as
        categories are selected.

        Once a non-category selection is made, the user's selection will be passed to
        the given callable, both as a string and as an index number. The index is given
        to ensure every selection has a unique identifier, so that selections with the
        same key in different categories can be distinguished between.

        The menus called by this function are not persistent and cannot perform
        complicated tasks like prompt for arbitrary input or jump multiple category
        levels at once - you'll have to use EvMenu itself if you want to take full
        advantage of its features.
    """

    # Pass kwargs to store data needed in the menu
    kwargs = {
        "index": index,
        "mark_category": mark_category,
        "go_back": go_back,
        "treestr": treestr,
        "callback": callback,
        "start_text": start_text,
    }

    # Initialize menu of selections
    evmenu.EvMenu(
        caller,
        "evennia.contrib.utils.tree_select",
        startnode="menunode_treeselect",
        startnode_input=None,
        cmd_on_exit=cmd_on_exit,
        **kwargs,
    )


def dashcount(entry):
    """
    Counts the number of dashes at the beginning of a string. This
    is needed to determine the depth of options in categories.

    Args:
        entry (str): String to count the dashes at the start of

    Returns:
        dashes (int): Number of dashes at the start
    """
    dashes = 0
    for char in entry:
        if char == "-":
            dashes += 1
        else:
            return dashes
    return dashes


def is_category(treestr, index):
    """
    Determines whether an option in a tree string is a category by
    whether or not there are additional options below it.

    Args:
        treestr (str): Multi-line string representing menu options
        index (int): Which line of the string to test

    Returns:
        is_category (bool): Whether the option is a category
    """
    opt_list = treestr.split("\n")
    # Not a category if it's the last one in the list
    if index == len(opt_list) - 1:
        return False
    # Not a category if next option is not one level deeper
    return not bool(dashcount(opt_list[index + 1]) != dashcount(opt_list[index]) + 1)


def parse_opts(treestr, category_index=None):
    """
    Parses a tree string and given index into a list of options. If
    category_index is none, returns all the options at the top level of
    the menu. If category_index corresponds to a category, returns a list
    of options under that category. If category_index corresponds to
    an option that is not a category, it's a selection and returns True.

    Args:
        treestr (str): Multi-line string representing menu options
        category_index (int): Index of category or None for top level

    Returns:
        kept_opts (list or True): Either a list of options in the selected
                                  category or True if a selection was made
    """
    dash_depth = 0
    opt_list = treestr.split("\n")
    kept_opts = []

    # If a category index is given
    if category_index != None:
        # If given index is not a category, it's a selection - return True.
        if not is_category(treestr, category_index):
            return True
        # Otherwise, change the dash depth to match the new category.
        dash_depth = dashcount(opt_list[category_index]) + 1
        # Delete everything before the category index
        opt_list = opt_list[category_index + 1 :]

    # Keep every option (referenced by index) at the appropriate depth
    cur_index = 0
    for option in opt_list:
        if dashcount(option) == dash_depth:
            if category_index == None:
                kept_opts.append((cur_index, option[dash_depth:]))
            else:
                kept_opts.append((cur_index + category_index + 1, option[dash_depth:]))
        # Exits the loop if leaving a category
        if dashcount(option) < dash_depth:
            return kept_opts
        cur_index += 1
    return kept_opts


def index_to_selection(treestr, index, desc=False):
    """
    Given a menu tree string and an index, returns the corresponding selection's
    name as a string. If 'desc' is set to True, will return the selection's
    description as a string instead.

    Args:
        treestr (str): Multi-line string representing menu options
        index (int): Index to convert to selection key or description

    Options:
        desc (bool): If true, returns description instead of key

    Returns:
        selection (str): Selection key or description if 'desc' is set
    """
    opt_list = treestr.split("\n")
    # Fetch the given line
    selection = opt_list[index]
    # Strip out the dashes at the start
    selection = selection[dashcount(selection) :]
    # Separate out description, if any
    if ":" in selection:
        # Split string into key and description
        selection = selection.split(":", 1)
        selection[1] = selection[1].strip(" ")
    else:
        # If no description given, set description to None
        selection = [selection, None]
    if not desc:
        return selection[0]
    else:
        return selection[1]


def go_up_one_category(treestr, index):
    """
    Given a menu tree string and an index, returns the category that the given option
    belongs to. Used for the 'go back' option.

    Args:
        treestr (str): Multi-line string representing menu options
        index (int): Index to determine the parent category of

    Returns:
        parent_category (int): Index of parent category
    """
    opt_list = treestr.split("\n")
    # Get the number of dashes deep the given index is
    dash_level = dashcount(opt_list[index])
    # Delete everything after the current index
    opt_list = opt_list[: index + 1]

    # If there's no dash, return 'None' to return to base menu
    if dash_level == 0:
        return None
    current_index = index
    # Go up through each option until we find one that's one category above
    for selection in reversed(opt_list):
        if dashcount(selection) == dash_level - 1:
            return current_index
        current_index -= 1


def optlist_to_menuoptions(treestr, optlist, index, mark_category, go_back):
    """
    Takes a list of options processed by parse_opts and turns it into
    a list/dictionary of menu options for use in menunode_treeselect.

    Args:
        treestr (str): Multi-line string representing menu options
        optlist (list): List of options to convert to EvMenu's option format
        index (int): Index of current category
        mark_category (bool): Whether or not to mark categories with [+]
        go_back (bool): Whether or not to add an option to go back in the menu

    Returns:
        menuoptions (list of dicts): List of menu options formatted for use
            in EvMenu, each passing a different "newindex" kwarg that changes
            the menu level or makes a selection
    """

    menuoptions = []
    cur_index = 0
    for option in optlist:
        index_to_add = optlist[cur_index][0]
        menuitem = {}
        keystr = index_to_selection(treestr, index_to_add)
        if mark_category and is_category(treestr, index_to_add):
            # Add the [+] to the key if marking categories, and the key by itself as an alias
            menuitem["key"] = [keystr + " [+]", keystr]
        else:
            menuitem["key"] = keystr
        # Get the option's description
        desc = index_to_selection(treestr, index_to_add, desc=True)
        if desc:
            menuitem["desc"] = desc
        # Passing 'newindex' as a kwarg to the node is how we move through the menu!
        menuitem["goto"] = ["menunode_treeselect", {"newindex": index_to_add}]
        menuoptions.append(menuitem)
        cur_index += 1
    # Add option to go back, if needed
    if index != None and go_back == True:
        gobackitem = {
            "key": ["<< Go Back", "go back", "back"],
            "desc": "Return to the previous menu.",
            "goto": ["menunode_treeselect", {"newindex": go_up_one_category(treestr, index)}],
        }
        menuoptions.append(gobackitem)
    return menuoptions


def menunode_treeselect(caller, raw_string, **kwargs):
    """
    This is the repeating menu node that handles the tree selection.
    """

    # If 'newindex' is in the kwargs, change the stored index.
    if "newindex" in kwargs:
        caller.ndb._menutree.index = kwargs["newindex"]

    # Retrieve menu info
    index = caller.ndb._menutree.index
    mark_category = caller.ndb._menutree.mark_category
    go_back = caller.ndb._menutree.go_back
    treestr = caller.ndb._menutree.treestr
    callback = caller.ndb._menutree.callback
    start_text = caller.ndb._menutree.start_text

    # List of options if index is 'None' or category, or 'True' if a selection
    optlist = parse_opts(treestr, category_index=index)

    # If given index returns optlist as 'True', it's a selection. Pass to callback and end the menu.
    if optlist == True:
        selection = index_to_selection(treestr, index)
        try:
            callback(caller, treestr, index, selection)
        except Exception:
            log_trace("Error in tree selection callback.")

        # Returning None, None ends the menu.
        return None, None

    # Otherwise, convert optlist to a list of menu options.
    else:
        options = optlist_to_menuoptions(treestr, optlist, index, mark_category, go_back)
        if index == None:
            # Use start_text for the menu text on the top level
            text = start_text
        else:
            # Use the category name and description (if any) as the menu text
            if index_to_selection(treestr, index, desc=True) != None:
                text = (
                    "|w"
                    + index_to_selection(treestr, index)
                    + "|n: "
                    + index_to_selection(treestr, index, desc=True)
                )
            else:
                text = "|w" + index_to_selection(treestr, index) + "|n"
        return text, options


# The rest of this module is for the example menu and command! It'll change the color of your name.

"""
Here's an example string that you can initialize a menu from. Note the dashes at
the beginning of each line - that's how menu option depth and hierarchy is determined.
"""

NAMECOLOR_MENU = """Set name color: Choose a color for your name!
-Red shades: Various shades of |511red|n
--Red: |511Set your name to Red|n
--Pink: |533Set your name to Pink|n
--Maroon: |301Set your name to Maroon|n
-Orange shades: Various shades of |531orange|n
--Orange: |531Set your name to Orange|n
--Brown: |321Set your name to Brown|n
--Sienna: |420Set your name to Sienna|n
-Yellow shades: Various shades of |551yellow|n
--Yellow: |551Set your name to Yellow|n
--Gold: |540Set your name to Gold|n
--Dandelion: |553Set your name to Dandelion|n
-Green shades: Various shades of |141green|n
--Green: |141Set your name to Green|n
--Lime: |350Set your name to Lime|n
--Forest: |032Set your name to Forest|n
-Blue shades: Various shades of |115blue|n
--Blue: |115Set your name to Blue|n
--Cyan: |155Set your name to Cyan|n
--Navy: |113Set your name to Navy|n
-Purple shades: Various shades of |415purple|n
--Purple: |415Set your name to Purple|n
--Lavender: |535Set your name to Lavender|n
--Fuchsia: |503Set your name to Fuchsia|n
Remove name color: Remove your name color, if any"""


class CmdNameColor(Command):
    """
    Set or remove a special color on your name. Just an example for the
    easy menu selection tree contrib.
    """

    key = "namecolor"

    def func(self):
        # This is all you have to do to initialize a menu!
        init_tree_selection(
            NAMECOLOR_MENU, self.caller, change_name_color, start_text="Name color options:"
        )


def change_name_color(caller, treestr, index, selection):
    """
    Changes a player's name color.

    Args:
        caller (obj): Character whose name to color.
        treestr (str): String for the color change menu - unused
        index (int): Index of menu selection - unused
        selection (str): Selection made from the name color menu - used
            to determine the color the player chose.
    """

    # Store the caller's uncolored name
    if not caller.db.uncolored_name:
        caller.db.uncolored_name = caller.key

    # Dictionary matching color selection names to color codes
    colordict = {
        "Red": "|511",
        "Pink": "|533",
        "Maroon": "|301",
        "Orange": "|531",
        "Brown": "|321",
        "Sienna": "|420",
        "Yellow": "|551",
        "Gold": "|540",
        "Dandelion": "|553",
        "Green": "|141",
        "Lime": "|350",
        "Forest": "|032",
        "Blue": "|115",
        "Cyan": "|155",
        "Navy": "|113",
        "Purple": "|415",
        "Lavender": "|535",
        "Fuchsia": "|503",
    }

    # I know this probably isn't the best way to do this. It's just an example!
    if selection == "Remove name color":  # Player chose to remove their name color
        caller.key = caller.db.uncolored_name
        caller.msg("Name color removed.")
    elif selection in colordict:
        newcolor = colordict[selection]  # Retrieve color code based on menu selection
        caller.key = newcolor + caller.db.uncolored_name + "|n"  # Add color code to caller's name
        caller.msg(newcolor + ("Name color changed to %s!" % selection) + "|n")



"""
Test wilderness

"""

from evennia import DefaultCharacter
from evennia.utils.create import create_object
from evennia.utils.test_resources import BaseEvenniaTest

from . import wilderness


class TestWilderness(BaseEvenniaTest):
    def setUp(self):
        super().setUp()
        self.char1 = create_object(DefaultCharacter, key="char1")
        self.char2 = create_object(DefaultCharacter, key="char2")

    def get_wilderness_script(self, name="default"):
        w = wilderness.WildernessScript.objects.get("default")
        return w

    def test_create_wilderness_default_name(self):
        wilderness.create_wilderness()
        w = self.get_wilderness_script()
        self.assertIsNotNone(w)

    def test_create_wilderness_custom_name(self):
        name = "customname"
        wilderness.create_wilderness(name)
        w = self.get_wilderness_script(name)
        self.assertIsNotNone(w)

    def test_enter_wilderness(self):
        wilderness.create_wilderness()
        wilderness.enter_wilderness(self.char1)
        self.assertIsInstance(self.char1.location, wilderness.WildernessRoom)
        w = self.get_wilderness_script()
        self.assertEqual(w.itemcoordinates[self.char1], (0, 0))

    def test_enter_wilderness_custom_coordinates(self):
        wilderness.create_wilderness()
        wilderness.enter_wilderness(self.char1, coordinates=(1, 2))
        self.assertIsInstance(self.char1.location, wilderness.WildernessRoom)
        w = self.get_wilderness_script()
        self.assertEqual(w.itemcoordinates[self.char1], (1, 2))

    def test_enter_wilderness_custom_name(self):
        name = "customnname"
        wilderness.create_wilderness(name)
        wilderness.enter_wilderness(self.char1, name=name)
        self.assertIsInstance(self.char1.location, wilderness.WildernessRoom)

    def test_wilderness_correct_exits(self):
        wilderness.create_wilderness()
        wilderness.enter_wilderness(self.char1)

        # By default we enter at a corner (0, 0), so only a few exits should
        # be visible / traversable
        exits = [
            i
            for i in self.char1.location.contents
            if i.destination and (i.access(self.char1, "view") or i.access(self.char1, "traverse"))
        ]

        self.assertEqual(len(exits), 3)
        exitsok = ["north", "northeast", "east"]
        for each_exit in exitsok:
            self.assertTrue(any([e for e in exits if e.key == each_exit]))

        # If we move to another location not on an edge, then all directions
        # should be visible / traversable
        wilderness.enter_wilderness(self.char1, coordinates=(1, 1))
        exits = [
            i
            for i in self.char1.location.contents
            if i.destination and (i.access(self.char1, "view") or i.access(self.char1, "traverse"))
        ]
        self.assertEqual(len(exits), 8)
        exitsok = [
            "north",
            "northeast",
            "east",
            "southeast",
            "south",
            "southwest",
            "west",
            "northwest",
        ]
        for each_exit in exitsok:
            self.assertTrue(any([e for e in exits if e.key == each_exit]))

    def test_room_creation(self):
        # Pretend that both char1 and char2 are connected...
        self.char1.sessions.add(1)
        self.char2.sessions.add(1)
        self.assertTrue(self.char1.has_account)
        self.assertTrue(self.char2.has_account)

        wilderness.create_wilderness()
        w = self.get_wilderness_script()

        # We should have no unused room after moving the first account in.
        self.assertEqual(len(w.db.unused_rooms), 0)
        w.move_obj(self.char1, (0, 0))
        self.assertEqual(len(w.db.unused_rooms), 0)
        # and there should be one used room
        self.assertEqual(len(w.db.rooms.values()), 1)

        # And also no unused room after moving the second one in.
        w.move_obj(self.char2, (1, 1))
        self.assertEqual(len(w.db.unused_rooms), 0)
        # but now there should be 2 used rooms
        self.assertEqual(len(w.db.rooms.values()), 2)

        # Now, if char2 moves into char1's room, we should have one unused room
        # Which should be char2's old room that got created.
        w.move_obj(self.char2, (0, 0))
        self.assertEqual(len(w.db.unused_rooms), 1)
        self.assertEqual(self.char1.location, self.char2.location)

        # And if char2 moves back out, that unused room should be put back to
        # use again.
        w.move_obj(self.char2, (1, 1))
        self.assertNotEqual(self.char1.location, self.char2.location)
        self.assertEqual(len(w.db.unused_rooms), 0)

    def test_get_new_coordinates(self):
        loc = (1, 1)
        directions = {
            "north": (1, 2),
            "northeast": (2, 2),
            "east": (2, 1),
            "southeast": (2, 0),
            "south": (1, 0),
            "southwest": (0, 0),
            "west": (0, 1),
            "northwest": (0, 2),
        }
        for direction, correct_loc in directions.items():
            new_loc = wilderness.get_new_coordinates(loc, direction)
            self.assertEqual(new_loc, correct_loc, direction)

    def test_preserve_items(self):
        wilderness.create_wilderness()
        w = self.get_wilderness_script()

        # move char and obj to wilderness
        wilderness.enter_wilderness(self.char1)
        wilderness.enter_wilderness(self.obj1)

        # move to a new room
        w.move_obj(self.char1, (1, 1))
        # the room should be remapped and 0,0 should not exist
        self.assertTrue((0, 0) not in w.db.rooms)
        self.assertEqual(1, len(w.db.rooms))
        # verify obj1 moved to None
        self.assertIsNone(self.obj1.location)

        # now change to preserve items
        w.preserve_items = True
        wilderness.enter_wilderness(self.obj1, (1, 1))
        # move the character again
        w.move_obj(self.char1, (0, 1))
        # check that the previous room was preserved
        self.assertIn((1, 1), w.db.rooms)
        self.assertEqual(2, len(w.db.rooms))
        # and verify that obj1 is still at 1,1
        self.assertEqual(self.obj1.location, w.db.rooms[(1, 1)])
